/*  NAME:
        Collier_LaplacianFilter.cpp
 
    DESCRIPTION:
        LaPlacianFilter class implementation.
 
    COPYRIGHT:
        csc583VisionLib Copyright  2005, csc583VisionLib Developers.
  
        This library is free software; you can redistribute it and/or
        modify it under the terms of the GNU Lesser General Public
        License as published by the Free Software Foundation; either
        version 2 of the License, or (at your option) any later version.
 
        This library is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
        Lesser General Public License for more details.
 
        You should have received a copy of the GNU Lesser General Public
        License along with this library; if not, write to the Free Software
        Foundation Inc, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
    ___________________________________________________________________________
*/

#include <math.h>
//
#include "RasterImage.h"
#include "RasterImage_gray.h"
#include "RasterImage_RGB.h"
#include "RasterImage_RGBa.h"
#include "RasterImage_gray_F.h"

#include "collier_CSC583LIB.h"
#include "collier_ImageMath.h"
#include "collier_ImageLaplacian_Sobel.h"


using namespace csc583Base;
using namespace csc583ImageProcessing;


//----------------------------------------------------------
//  static members
//----------------------------------------------------------
// version from web page
int nLaPlacianCoeffs1[] = {0,  1, 0,
                          1, -4, 1,
                          0,  1, 0};
// version from notes
int nLaPlacianCoeffs2[] = {1,  1, 1,
                          1, -8, 1,
                          1,  1, 1};

//----------------------------------------------------------
//  Constructors and destructors
//----------------------------------------------------------

ImageLaplacian_Sobel::ImageLaplacian_Sobel(short wVersion) : ImageOperator()
{ 
	fOutRaster_ = NULL;

	isDoubleOperator_ = TRUE;
	fFilterRaster_ = new RasterImage_gray_F(3, 3);
	float *fFilterArray = fFilterRaster_->getGrayF();
	
 	if (wVersion == 1)
 	{
 		for (int i=0; i<9; i++)
 		{
 			fFilterArray[i] = ((float)nLaPlacianCoeffs1[i])/8.0f;
 		}
 	}
 	else
 	{
 		for (int i=0; i<9; i++)
 		{
 			fFilterArray[i] = ((float)nLaPlacianCoeffs2[i])/16.0f;
 		}
 	}
	
} // end LaPlacianFilter()


ImageLaplacian_Sobel::~ImageLaplacian_Sobel()
{
	// clean up our temp output spaces if they were allocated
	if (fOutRaster_ != NULL)
	{
		delete []fOutRaster_;
	}

} // end LaPlacianFilter~()

//----------------------------------------------------------
//  application functions
//----------------------------------------------------------


RasterImage *ImageLaplacian_Sobel::applyTo(RasterImage *imgIn)
{
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**
	//
	// !!!!! CAVEAT :: USER BEWARE !!!!!
	//
	// Returning a pointer from a function other than "new"
	// 	is never considered a good thing because of the possible
	//	risk of easily generating memory leaks. Therefore,
	//
	// THIS FUNCTION RETURNS A POINTER TO A PROTECTED VARIABLE
	//
	// Each call to this function will delete any prior results
	//	and reinit the internal variable.
	//
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**
	
	return(applyTo(imgIn, imgIn->getValidRect()));

} // end ImageLaplacian_Sobel.applyTo(RasterImage *imgIn)




RasterImage *ImageLaplacian_Sobel::applyTo(RasterImage *imgIn, ImageRect *theRect)
{
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**
	//
	// !!!!! CAVEAT :: USER BEWARE !!!!!
	//
	// Returning a pointer from a function other than "new"
	// 	is never considered a good thing because of the possible
	//	risk of easily generating memory leaks. Therefore,
	//
	// THIS FUNCTION RETURNS A POINTER TO A PROTECTED VARIABLE
	//
	// Each call to this function will delete any prior results
	//	and reinit the internal variable.
	//
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**

	// get a new raster to store the results into
	if (fOutRaster_ != NULL)
	{
		delete []fOutRaster_;
	}
	fOutRaster_ = new RasterImage_gray_F(imgIn);
	FAIL_CONDITION(fOutRaster_ == NULL,
                kRasterAllocationFailure,
                "Error allocating output raster in LaPlacianFilter::applyInto()");

	// apply the filter
	applyInto(imgIn, theRect, fOutRaster_);

	return(fOutRaster_);
	
} // end ImageLaplacian_Sobel.applyTo(RasterImage *imgIn, ImageRect *theRect)




void ImageLaplacian_Sobel::applyInto(RasterImage *imgIn, RasterImage *imgOut)
{
	applyInto(imgIn, imgIn->getValidRect(), imgOut);

} // end ImageLaplacian_Sobel.applyInto(RasterImage *imgIn, RasterImage *imgOut)




void ImageLaplacian_Sobel::applyInto(RasterImage *imgIn, ImageRect *theRect, RasterImage *imgOut)
{
	ImageMath ImageFilter;
	
	// use theRect to limit the filter on the input
	// save old valid rect, restore at end
	ImageRect	*pRect 	= new ImageRect(imgIn->getValidRect());
	imgIn->setValidRect(theRect);

	// call generic 2D filter routine to do all the heavy lifting
	ImageFilter.convolve(imgIn,fFilterRaster_,imgOut);
	
	// restore imgIn valid rect and delete our temp space
	imgIn->setValidRect(pRect);
	if (pRect!=NULL) delete pRect;


} // end ImageLaplacian_Sobel.applyInto(RasterImage *imgIn, ImageRect *theRect, RasterImage *imgOut)
