/*  NAME:
        Collier_GaussianFilter.cpp
 
    DESCRIPTION:
        GaussianFilter class implementation.
 
    COPYRIGHT:
        csc583VisionLib Copyright  2005, csc583VisionLib Developers.
  
        This library is free software; you can redistribute it and/or
        modify it under the terms of the GNU Lesser General Public
        License as published by the Free Software Foundation; either
        version 2 of the License, or (at your option) any later version.
 
        This library is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
        Lesser General Public License for more details.
 
        You should have received a copy of the GNU Lesser General Public
        License along with this library; if not, write to the Free Software
        Foundation Inc, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
    ___________________________________________________________________________
*/

#include <math.h>
//
#include "RasterImage.h"
#include "RasterImage_gray.h"
#include "RasterImage_RGB.h"
#include "RasterImage_RGBa.h"
#include "RasterImage_gray_F.h"


#include "collier_CSC583LIB.h"
#include "collier_ImageMath.h"
#include "collier_GaussianFilter.h"


using namespace csc583Base;
using namespace csc583ImageProcessing;


//----------------------------------------------------------
//  static members
//----------------------------------------------------------


//----------------------------------------------------------
//  Constructors and destructors
//----------------------------------------------------------

GaussianFilter::GaussianFilter(double theScale) : ImageOperator()
{
	dSigma_ = theScale;
	wWidth_ = ceil((short)dSigma_)*5;
	wRadius_ = (wWidth_ - 1)/2;
	fNormFactor_ = 0.0;
	
	isDoubleOperator_ = TRUE;
	fFilterRaster_ = new RasterImage_gray_F(wWidth_, wWidth_);
	fFilterArray_ = fFilterRaster_->getGrayF2D();
	
	// fill in the gaussian filter array
	float fDenomFactor = 1.0/(2.0*dSigma_*dSigma_);
	for (int i=0; i<wWidth_; i++)
	{
		float fX2 = (i-wRadius_)*(i-wRadius_);
		for (int j=0; j<wWidth_; j++)
		{
			float fY2 = (j-wRadius_)*(j-wRadius_);
			fFilterArray_[i][j] = exp(-fDenomFactor*(fX2+fY2));
			fNormFactor_ += fFilterArray_[i][j];
		}
	}
	
	fNormFactor_ = 1.0 / fNormFactor_;

	for (int i=0; i<wWidth_; i++)
	{
		for (int j=0; j<wWidth_; j++)
		{
			fFilterArray_[i][j] *= fNormFactor_;
		}
	}
	
	fOutRaster_ = (RasterImage_gray_F *) NULL;	
	 
	
} // end GaussianFilter(double theScale)


GaussianFilter::GaussianFilter(short theWidth) : ImageOperator()
{
	wWidth_ = theWidth;
	dSigma_ = ((float)wWidth_)/5.0;
	wRadius_ = (wWidth_ - 1)/2;
	fNormFactor_ = 0;

	isDoubleOperator_ = TRUE;
	fFilterRaster_ = new RasterImage_gray_F(wWidth_, wWidth_);
	fFilterArray_ = fFilterRaster_->getGrayF2D();
	
	// fill in the gaussian filter array
	float fDenomFactor = 1.0/(2.0*dSigma_*dSigma_);
	for (int i=0; i<wWidth_; i++)
	{
		float fX2 = (i-wRadius_)*(i-wRadius_);
		for (int j=0; j<wWidth_; j++)
		{
			float fY2 = (j-wRadius_)*(j-wRadius_);
			fFilterArray_[i][j] = exp(-fDenomFactor*(fX2+fY2));
			fNormFactor_ += fFilterArray_[i][j];
		}
	}
	
	fNormFactor_ = 1.0 / fNormFactor_;
	
	for (int i=0; i<wWidth_; i++)
	{
		for (int j=0; j<wWidth_; j++)
		{
			fFilterArray_[i][j] *= fNormFactor_;
		}
	}

	fOutRaster_ = (RasterImage_gray_F *) NULL;	
	

} // end GaussianFilter(short theWidth)



GaussianFilter::~GaussianFilter()
{
	// clean up our temp output spaces if they were allocated
	if (fOutRaster_ != NULL)
	{
		delete fOutRaster_;
	}
	if (fFilterRaster_ != NULL)
	{
		delete fFilterRaster_;
	}


} // end GaussianFilter~()

//----------------------------------------------------------
//  application functions
//----------------------------------------------------------


RasterImage *GaussianFilter::applyTo(RasterImage *imgIn)
{
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**
	//
	// !!!!! CAVEAT :: USER BEWARE !!!!!
	//
	// Returning a pointer from a function other than "new"
	// 	is never considered a good thing because of the possible
	//	risk of easily generating memory leaks. Therefore,
	//
	// THIS FUNCTION RETURNS A POINTER TO A PROTECTED VARIABLE
	//
	// Each call to this function will delete any prior results
	//	and reinit the internal variable.
	//
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**
	
	return(applyTo(imgIn, imgIn->getValidRect()));

} // end GaussianFilter.applyTo(RasterImage *imgIn)




RasterImage *GaussianFilter::applyTo(RasterImage *imgIn, ImageRect *theRect)
{
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**
	//
	// !!!!! CAVEAT :: USER BEWARE !!!!!
	//
	// Returning a pointer from a function other than "new"
	// 	is never considered a good thing because of the possible
	//	risk of easily generating memory leaks. Therefore,
	//
	// THIS FUNCTION RETURNS A POINTER TO A PROTECTED VARIABLE
	//
	// Each call to this function will delete any prior results
	//	and reinit the internal variable.
	//
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**

	// get a new raster to store the results into
	if (fOutRaster_ != NULL)
	{
		delete []fOutRaster_;
	}
	fOutRaster_ = new RasterImage_gray_F(imgIn);
	FAIL_CONDITION(fOutRaster_ == NULL,
                kRasterAllocationFailure,
                "Error allocating output raster in GaussianFilter::applyInto()");

	// apply the filter
	applyInto(imgIn, theRect, fOutRaster_);

	return(fOutRaster_);
	
} // end GaussianFilter.applyTo(RasterImage *imgIn, ImageRect *theRect)




void GaussianFilter::applyInto(RasterImage *imgIn, RasterImage *imgOut)
{
	applyInto(imgIn, imgIn->getValidRect(), imgOut);

} // end GaussianFilter.applyInto(RasterImage *imgIn, RasterImage *imgOut)




void GaussianFilter::applyInto(RasterImage *imgIn, ImageRect *theRect, RasterImage *imgOut)
{
	ImageMath ImageFilter;
	
	// use theRect to limit the filter on the input
	// save old valid rect, restore at end
	ImageRect	*pRect 	= new ImageRect(imgIn->getValidRect());
	imgIn->setValidRect(theRect);
	
	// apply the gaussian filter to the input
	ImageFilter.convolve(imgIn, fFilterRaster_, imgOut);
	
	// restore imgIn valid rect and delete our temp space
	imgIn->setValidRect(pRect);
	if (pRect!=NULL) delete pRect;

	
} // end GaussianFilter.applyInto(RasterImage *imgIn, ImageRect *theRect, RasterImage *imgOut)
