#ifndef IMAGE_FFT_H
#define IMAGE_FFT_H

#include "ImageOperator.h"

using namespace csc583Base;

namespace csc583ImageProcessing
{

    class CSC583LIB_EXPORT ImageFFT : public ImageOperator
    {
        public:

            /** Defines an FFT operator with a specified square size.
             *
             *  @param  nSize    dimension of the 2D FFT, MUST BE A POWER OF 2
             */
            ImageFFT(int nSize);
            
            /** Destructor
             */
            virtual ~ImageFFT(void);

            
            /** Applies this operator to an image, returns the result as a RasterImage.
             *
             *  
             *
             *  @param  imgIn       the RasterImage to which the operator is applied
             *  @return             the RasterImage obtained by applying the operator to
             *                      imgIn. Since this FFT is done in place, the returned
             *                      pointer is just a copy of the input.
             */
             
            RasterImage *applyTo(RasterImage *imgIn);

            /** Applies this operator to an image  within the rectangle specified and
             *  returns the result as a RasterImage.
             *
             *  The RasterImage returned is of the "float" form of the input image.
             *  For example, if imgIn was RasterImage_Gray, then the image
             *  returned should be RasterImage_Gray_F.
             *
             *  @param  imgIn       the RasterImage to which the operator is applied
             *  @param  theRect     the rectangle within which to apply the filter
             *                      This parameter IS IGNORED since this operator is
             *                      set up to use the ENTIRE (bounding rect) of the image.
             *  @return             the RasterImage obtained by applying the operator to
             *                      imgIn. Since this FFT is done in place, the returned
             *                      pointer is just a copy of the input
             */
             
            RasterImage *applyTo(RasterImage *imgIn, ImageRect *theRect);

            /** Applies this operator to a RasterImage and writes the result into a second
             *  RasterImage received as parameter. This operator is slightly slower since 
             *  it must make a copy of the data into the output buffer first.
             *
             *  @param  imgIn       the RasterImage to which the operator is applied
 			 *  @param  nDirection  + for forward FFT, - for inverse FFT
             *  @param  imgOut      the RasterImage into which the output is written
             */ 
            void applyInto(RasterImage *imgIn, RasterImage *imgOut);
            
            /** Applies this operator to a RasterImage within the rectangle specified
             *  and writes the result into the RasterImage received as parameter. 
             *  This operator is slightly slower since it must make a copy of the data 
             *  into the output buffer first.
             *
             *  @param  imgIn       the RasterImage to which the operator is applied
 			 *  @param  nDirection  + for forward FFT, - for inverse FFT
             *  @param  theRect     the rectangle within which to apply the filter
             *                      This parameter IS IGNORED since this operator is
             *                      set up to use the ENTIRE (bounding rect) of the image.
             *  @param  imgOut      the RasterImage into which the output is written
             */ 
            void applyInto(RasterImage *imgIn, ImageRect *theRect, RasterImage *imgOut);

            
            /** Applies this operator to a RasterImage within the rectangle specified
             *  and writes the result into the RasterImage received as parameter. 
             *  This operator is slightly slower since it must make a copy of the data 
             *  into the output buffer first.
             *
             *  @param  imgIn       the RasterImage to which the operator is applied
 			 *  @param  nPaddedSize size to use for padded FFT to avoid circular FFT
             *  @param  imgOut      the RasterImage into which the output is written
             *  @return             the RasterImage obtained by applying the operator to
             *                      imgIn. Since this FFT is done in place, the returned
             *                      pointer is just a copy of the input.
             */
             
            void applyInto(RasterImage *imgIn, int nPaddedSize, RasterImage *imgOut);


            /** Sets direction for FFT
             */
            void setForward(void);


            /** Sets direction for FFT
             */
            void setInverse(void);

        protected:

        	float *pfFwdWeights_;  // weights for FFT of specified size.
        	float *pfInvWeights_;  // weights for FFT of specified size.
        	int   *pnFwdIp_;       // forward FFT twiddle factors???
        	int   *pnInvIp_;       // forward FFT twiddle factors???
        	int    nSize_;	       // declared size for this FFT
			// needed since applyTo() and applyInto() are pre-defined
			int    nDirection_;    // direction +=forward, -=inverse
								

    };
}
#endif  //  IMAGE_FFT_H