/*  NAME:
        Collier_ImageGradient_Gaussian.cpp
 
    DESCRIPTION:
        GaussianFilter class implementation.
 
    COPYRIGHT:
        csc583VisionLib Copyright  2005, csc583VisionLib Developers.
  
        This library is free software; you can redistribute it and/or
        modify it under the terms of the GNU Lesser General Public
        License as published by the Free Software Foundation; either
        version 2 of the License, or (at your option) any later version.
 
        This library is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
        Lesser General Public License for more details.
 
        You should have received a copy of the GNU Lesser General Public
        License along with this library; if not, write to the Free Software
        Foundation Inc, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
    ___________________________________________________________________________
*/

#include <math.h>
//
#include "RasterImage.h"
#include "RasterImage_gray.h"
#include "RasterImage_RGB.h"
#include "RasterImage_RGBa.h"
#include "RasterImage_gray_F.h"

#include "VectorField.h"
#include "VectorField_F.h"
#include "VectorField_I.h"

#include "collier_CSC583LIB.h"
#include "collier_ImageMath.h"
#include "collier_ImageGradient_Gaussian.h"

using namespace csc583Base;
using namespace csc583ImageProcessing;


//----------------------------------------------------------
//  static members
//----------------------------------------------------------

// d/dx matrix
static int nDX[] = {0,  0, 0,
                   -1, 0, 1,
                    0,  0, 0};

// d/dy matrix
static int nDY[] = {0, -1, 0,
                    0, 0, 0,
                    0,  1, 0};

//----------------------------------------------------------
//  Constructors and destructors
//----------------------------------------------------------


ImageGradient_Gaussian::ImageGradient_Gaussian(double theScale) : VectorOperator()
{
	float **fTempMatrix;
	float *fTempArray;

	//
	// set up gaussian filter
	//
	dSigma_ = theScale;
	wWidth_ = ceil((short)dSigma_)*5;
	wRadius_ = (wWidth_ - 1)/2;
	fNormFactor_ = 0.0;
	
	isDoubleOperator_ = TRUE;
	FilterRasterF_ = new RasterImage_gray_F(wWidth_, wWidth_);
	fTempMatrix = FilterRasterF_->getGrayF2D();
	
	// fill in the gaussian filter array
	float fDenomFactor = 1.0/(2.0*dSigma_*dSigma_);
	for (int i=0; i<wWidth_; i++)
	{
		float fX2 = (i-wRadius_)*(i-wRadius_);
		for (int j=0; j<wWidth_; j++)
		{
			float fY2 = (j-wRadius_)*(j-wRadius_);
			fTempMatrix[i][j] = exp(-fDenomFactor*(fX2+fY2));
			fNormFactor_ += fTempMatrix[i][j];
		}
	}
	
	fNormFactor_ = 1.0 / fNormFactor_;
	
	//
	// set up d/dx filter
	//
	DxRasterF_ = new RasterImage_gray_F(3, 3);
	fTempArray = DxRasterF_->getGrayF();
	
	for (int i=0; i<9; i++)
 	{
 		fTempArray[i] = (float)nDX[i]/2;
 	}

	//
	// set up d/dy filter
	//
	DyRasterF_ = new RasterImage_gray_F(3, 3);
	fTempArray = DyRasterF_->getGrayF();
	
	for (int i=0; i<9; i++)
 	{
 		fTempArray[i] = (float)nDY[i]/2;
 	}
	
	//
	// set internal ptrs to NULL for safety
	//
	GaussianRasterF_ = (RasterImage_gray_F *) NULL;
	X_RasterF_ = (RasterImage_gray_F *) NULL;
	Y_RasterF_ = (RasterImage_gray_F *) NULL;
	OutVectorFieldF_ = (VectorField_F *) NULL;
	
} // end ImageGradient_Gaussian::ImageGradient_Gaussian(double theScale)


ImageGradient_Gaussian::ImageGradient_Gaussian(short theWidth) : VectorOperator()
{
	float **fTempMatrix;
	float *fTempArray;
	
	//
	// set up gaussian filter
	//
	wWidth_ = theWidth;
	dSigma_ = ((float)wWidth_)/5.0;
	wRadius_ = (wWidth_ - 1)/2;
	fNormFactor_ = 0;

	isDoubleOperator_ = TRUE;
	FilterRasterF_ = new RasterImage_gray_F(wWidth_, wWidth_);
	fTempMatrix = FilterRasterF_->getGrayF2D();
	
	// fill in the gaussian filter array
	float fDenomFactor = 1.0/(2.0*dSigma_*dSigma_);
	for (int i=0; i<wWidth_; i++)
	{
		float fX2 = (i-wRadius_)*(i-wRadius_);
		for (int j=0; j<wWidth_; j++)
		{
			float fY2 = (j-wRadius_)*(j-wRadius_);
			fTempMatrix[i][j] = exp(-fDenomFactor*(fX2+fY2));
			fNormFactor_ += fTempMatrix[i][j];
		}
	}
	
	fNormFactor_ = 1.0 / fNormFactor_;
	
	//
	// set up d/dx filter
	//
	DxRasterF_ = new RasterImage_gray_F(3, 3);
	fTempArray = DxRasterF_->getGrayF();
	
	for (int i=0; i<9; i++)
 	{
 		fTempArray[i] = (float)nDX[i]/2;
 	}

	//
	// set up d/dy filter
	//
	DyRasterF_ = new RasterImage_gray_F(3, 3);
	fTempArray = DyRasterF_->getGrayF();
	
	for (int i=0; i<9; i++)
 	{
 		fTempArray[i] = (float)nDY[i]/2;
 	}
	
	//
	// set internal ptrs to NULL for safety
	//
	GaussianRasterF_ 	= (RasterImage_gray_F *) NULL;
	X_RasterF_ 			= (RasterImage_gray_F *) NULL;
	Y_RasterF_			= (RasterImage_gray_F *) NULL;
	OutVectorFieldF_ 	= (VectorField_F *) NULL;

} // end ImageGradient_Gaussian::ImageGradient_Gaussian(short theWidth)



ImageGradient_Gaussian::~ImageGradient_Gaussian()
{
	// clean up our temp output spaces if they were allocated
	if (DxRasterF_ 			!= NULL)	delete DxRasterF_;
	if (DyRasterF_ 			!= NULL)	delete DyRasterF_;
	if (FilterRasterF_ 		!= NULL)	delete FilterRasterF_;
	if (GaussianRasterF_ 	!= NULL)	delete GaussianRasterF_;
	if (X_RasterF_ 			!= NULL)	delete X_RasterF_;
	if (Y_RasterF_ 			!= NULL)	delete Y_RasterF_;
	if (OutVectorFieldF_ 	!= NULL)	delete OutVectorFieldF_;


} // end ~ImageGradient_Gaussian()

//----------------------------------------------------------
//  application functions
//----------------------------------------------------------


VectorField *ImageGradient_Gaussian::applyTo(RasterImage *imgIn)
{
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**
	//
	// !!!!! CAVEAT :: USER BEWARE !!!!!
	//
	// Returning a pointer from a function other than "new"
	// 	is never considered a good thing because of the possible
	//	risk of easily generating memory leaks. Therefore,
	//
	// THIS FUNCTION RETURNS A POINTER TO A PROTECTED VARIABLE
	//
	// Each call to this function will delete any prior results
	//	and reinit the internal variable.
	//
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**
	
	return(applyTo(imgIn, imgIn->getValidRect()));

} // end ImageGradient_Gaussian.applyTo(RasterImage *imgIn)




VectorField *ImageGradient_Gaussian::applyTo(RasterImage *imgIn, ImageRect *theRect)
{
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**
	//
	// !!!!! CAVEAT :: USER BEWARE !!!!!
	//
	// Returning a pointer from a function other than "new"
	// 	is never considered a good thing because of the possible
	//	risk of easily generating memory leaks. Therefore,
	//
	// THIS FUNCTION RETURNS A POINTER TO A PROTECTED VARIABLE
	//
	// Each call to this function will delete any prior results
	//	and reinit the internal variable.
	//
	//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**

	// get a new raster to store the results into
	if (OutVectorFieldF_ != NULL)
	{
		delete OutVectorFieldF_;
	}
	OutVectorFieldF_ = new VectorField_F(imgIn->getBoundRect());
		FAIL_CONDITION(OutVectorFieldF_ == NULL,
                kRasterAllocationFailure,
                "Error allocating output vector in ImageGradient_Gaussian::applyInto()");

	// apply the filter
	applyInto(imgIn, theRect, OutVectorFieldF_);

	return((VectorField *)OutVectorFieldF_);
	
} // end ImageGradient_Gaussian.applyTo(RasterImage *imgIn, ImageRect *theRect)




void ImageGradient_Gaussian::applyInto(RasterImage *imgIn, VectorField *vectOut)
{
	applyInto(imgIn, imgIn->getValidRect(), vectOut);

} // end ImageGradient_Gaussian.applyInto(RasterImage *imgIn, VectorField *vectOut)


void ImageGradient_Gaussian::applyInto(RasterImage *imgIn, ImageRect *theRect, 
											VectorField *vectOut)
{
	ImageMath ImageFilter;

	// get temp rasters
	if (GaussianRasterF_ != NULL)
	{
		delete GaussianRasterF_;
	}
	GaussianRasterF_ = new RasterImage_gray_F(imgIn);
	FAIL_CONDITION(GaussianRasterF_ == NULL,
                kRasterAllocationFailure,
                "Error allocating gaussian raster in ImageGradient_Gaussian::applyInto()");
	if (X_RasterF_ != NULL)
	{
		delete X_RasterF_;
	}
	X_RasterF_ = new RasterImage_gray_F(imgIn);
	FAIL_CONDITION(X_RasterF_ == NULL,
                kRasterAllocationFailure,
                "Error allocating X raster in ImageGradient_Gaussian::applyInto()");
	if (Y_RasterF_ != NULL)
	{
		delete Y_RasterF_;
	}
	Y_RasterF_ = new RasterImage_gray_F(imgIn);
	FAIL_CONDITION(Y_RasterF_ == NULL,
                kRasterAllocationFailure,
                "Error allocating Y raster in ImageGradient_Gaussian::applyInto()");

	// use theRect to limit the filter on the input
	// save old valid rect, restore at end
	ImageRect	*pRect 	= new ImageRect(imgIn->getValidRect());
	imgIn->setValidRect(theRect);
	
	// apply the gaussian filter to the input
	ImageFilter.convolve(imgIn,FilterRasterF_,GaussianRasterF_);

	// apply the D/DX filter to the gaussian filtered input
	ImageFilter.convolve(GaussianRasterF_,DxRasterF_,X_RasterF_);
	
	// apply the D/DY filter to the gaussian filtered input
	ImageFilter.convolve(GaussianRasterF_,DyRasterF_,Y_RasterF_);
					  
	// restore imgIn valid rect and delete our temp space
	imgIn->setValidRect(pRect);
	if (pRect!=NULL) delete pRect;
		  
	// set gradient into vector space				  
	float **fXvect = (float **)vectOut->getRasterX2D();
	float **fYvect = (float **)vectOut->getRasterY2D();
	float **fDX    = X_RasterF_->getGrayF2D();
	float **fDY    = Y_RasterF_->getGrayF2D();

	for (int i=theRect->getTop(); i<=theRect->getBottom(); i++)
	{
		for (int j=theRect->getLeft(); j<=theRect->getRight(); j++)
		{
			fXvect[i][j] = fDX[i][j];
			fYvect[i][j] = fDY[i][j];
		}
	}
	
	// set the valid rect of the output vector
	vectOut->setValidRect(theRect->getLeft() + wRadius_+1,	 // start x
						  theRect->getTop() + wRadius_ +1,      // start y
						  theRect->getWidth() - 2*wRadius_ - 2,	 // width
						  theRect->getHeight() - 2*wRadius_ - 2);	 // height

	

} // end ImageGradient_Gaussian.applyInto(RasterImage *imgIn, ImageRect *theRect, VectorField *vectOut)

