/*  NAME:
        Collier_ImageMath.cpp
 
    DESCRIPTION:
        Basic Image Operator class implementation.
 
    COPYRIGHT:
        csc583VisionLib Copyright  2005, csc583VisionLib Developers.
  
        This library is free software; you can redistribute it and/or
        modify it under the terms of the GNU Lesser General Public
        License as published by the Free Software Foundation; either
        version 2 of the License, or (at your option) any later version.
 
        This library is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
        Lesser General Public License for more details.
 
        You should have received a copy of the GNU Lesser General Public
        License along with this library; if not, write to the Free Software
        Foundation Inc, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
    ___________________________________________________________________________
*/

#include <math.h>
#include <limits.h>
#include <time.h>

#include "csc583Lib.h"
//
#include "RasterImage.h"
#include "RasterImage_gray.h"
#include "RasterImage_RGB.h"
#include "RasterImage_RGBa.h"
#include "RasterImage_gray_F.h"

#include "collier_CSC583LIB.h"
#include "collier_ImageMath.h"

using namespace csc583Base;
using namespace csc583ImageProcessing;


//----------------------------------------------------------
//  static members
//----------------------------------------------------------


//----------------------------------------------------------
//  Constructors and destructors
//----------------------------------------------------------

ImageMath::ImageMath(void)
{
} // end ImageMath(void)


ImageMath::~ImageMath()
{
} // end ~ImageMath()

//----------------------------------------------------------
//  application functions
//----------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////////
//
// COPY functions

void ImageMath::copy(RasterImage *imgIn, RasterImage *imgOut)
{
	ImageRect *inRect  = imgIn->getValidRect(),
			  *outRect = imgOut->getBoundRect();

	int nInXo  = inRect->getLeft(), 
		nInXn  = inRect->getRight(), 
	    nInYo  = inRect->getTop(), 
		nInYn  = inRect->getBottom(); 
		
	// make sure output is larger than input
    FAIL_CONDITION(imgOut->getHeight() < inRect->getHeight(),
                kRasterImageError,
                "Output image has less rows than inputs in copy(In,Out)");
    FAIL_CONDITION(imgOut->getWidth() < inRect->getWidth(),
                kRasterImageError,
                "Output image has less columns than inputs in copy(In,Out)");
	
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgIn->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input rasters accepted in copy(In,Out)");
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in copy(In,Out)");
                
	// for both floats
	 if (imgIn->hasFloatRaster() && imgOut->hasFloatRaster())
	 {
	 	float** fIn 	= ((RasterImage_gray_F*)imgIn)->getGrayF2D();
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		
		// just use imgIn limits since both are the same
		for (int nYi = nInYo; nYi <= nInYn; nYi++)
		{
			for (int nXi = nInXo; nXi <= nInXn; nXi++)
			{
				fOut[nYi][nXi] = fIn[nYi][nXi];
			}
		}	
	 }
	 // for both unsigned chars
	 else if (!imgIn->hasFloatRaster() && !imgOut->hasFloatRaster())
	 {
	 	unsigned char** ubIn 	= ((RasterImage_gray*)imgIn)->getGray2D();
		unsigned char** ubOut 	= ((RasterImage_gray*)imgOut)->getGray2D();
		
		// just use imgIn limits since both are the same
		for (int nYi = nInYo; nYi <= nInYn; nYi++)
		{
			for (int nXi = nInXo; nXi <= nInXn; nXi++)
			{
				ubOut[nYi][nXi] = ubIn[nYi][nXi];
			}
		}	
	 }
	 else
	 {
	 	// mismatched types, FAIL for now
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same typed rasters accepted in copy(In,Out)");
     }

     // now set the output valid rectangle
     imgOut->setValidRect(imgIn->getValidRect());
	 	
} // end ImageMath::copy(RasterImage *imgIn, RasterImage *imgOut)


//-------------------------------

void ImageMath::copyTo(RasterImage *imgIn, int nXo, int nYo, RasterImage *imgOut)
{
	ImageRect *inRect  = imgIn->getValidRect(),
			  *outRect = imgOut->getBoundRect();

	int nInXo  = inRect->getLeft(), 
		nInXn  = inRect->getRight(), 
	    nInYo  = inRect->getTop(), 
		nInYn  = inRect->getBottom(); 

	int i, j;
		
	// make sure output is larger than input valid rect + origin
    FAIL_CONDITION(imgOut->getHeight() < (nYo + inRect->getHeight()),
                kRasterImageError,
                "Output image has less rows than needed in copyTo(In,X,Y, Out)");
    FAIL_CONDITION(imgOut->getWidth() < (nXo + inRect->getWidth()),
                kRasterImageError,
                "Output image has less columns than needed in copyTo(In,X,Y, Out)");
	
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgIn->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input rasters accepted in copy(In,Out)");
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in copy(In,Out)");
                
	// for both floats
	 if (imgIn->hasFloatRaster() && imgOut->hasFloatRaster())
	 {
	 	float** fIn 	= ((RasterImage_gray_F*)imgIn)->getGrayF2D();
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		    
		// just use imgIn limits since both are the same
		i=0;
		for (int nYi = nInYo; nYi <= nInYn; nYi++)
		{
			j=0;
			for (int nXi = nInXo; nXi <= nInXn; nXi++)
			{
				fOut[i+nYo][j+nXo] = fIn[nYi][nXi];
				j++;
			}
			i++;
		}	
	 }
	 // for both unsigned chars
	 else if (!imgIn->hasFloatRaster() && !imgOut->hasFloatRaster())
	 {
	 	unsigned char** ubIn 	= ((RasterImage_gray*)imgIn)->getGray2D();
		unsigned char** ubOut 	= ((RasterImage_gray*)imgOut)->getGray2D();
		
		// just use imgIn limits since both are the same
		i=0;
		for (int nYi = nInYo; nYi <= nInYn; nYi++)
		{
			j=0;
			for (int nXi = nInXo; nXi <= nInXn; nXi++)
			{
				ubOut[i+nYo][j+nXo] = ubIn[nYi][nXi];
				j++;
			}
			i++;
		}	
	 }
	 else
	 {
	 	// mismatched types, FAIL for now
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same typed rasters accepted in copy(In,Out)");
     }

     // now set the output valid rectangle
     imgOut->setValidRect(nXo,nYo,inRect->getWidth(),inRect->getHeight());
	 	
} // end ImageMath::copyTo(RasterImage *imgIn, int nXo, int nYo, RasterImage *imgOut)


//-------------------------------

void ImageMath::copyFlip(RasterImage *imgIn, RasterImage *imgOut)
{

	ImageStats_Type ImageStats;
	
	ImageRect *inRect  = imgIn->getValidRect(),
			  *outRect = imgOut->getBoundRect();

	int nInXo  = inRect->getLeft(), 
		nInXn  = inRect->getRight(), 
	    nInYo  = inRect->getTop(), 
		nInYn  = inRect->getBottom(); 
		
	// make sure output is larger than input
    FAIL_CONDITION(imgOut->getHeight() < imgIn->getHeight(),
                kRasterImageError,
                "Output image has less rows than inputs in copy(In,Out)");
    FAIL_CONDITION(imgOut->getWidth() < imgIn->getWidth(),
                kRasterImageError,
                "Output image has less columns than inputs in copy(In,Out)");
	
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgIn->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input rasters accepted in copy(In,Out)");
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in copy(In,Out)");

	// get min and max of input image
    getStats(imgIn,&ImageStats);
                
	// for both floats
	 if (imgIn->hasFloatRaster() && imgOut->hasFloatRaster())
	 {
	 	float** fIn 	= ((RasterImage_gray_F*)imgIn)->getGrayF2D();
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		
		// just use imgIn limits since both are the same
		for (int nYi = nInYo; nYi <= nInYn; nYi++)
		{
			for (int nXi = nInXo; nXi <= nInXn; nXi++)
			{
				fOut[nYi][nXi] = ImageStats.fMax - (fIn[nYi][nXi] - ImageStats.fMin);
			}
		}	
	 }
	 // for both unsigned chars
	 else if (!imgIn->hasFloatRaster() && !imgOut->hasFloatRaster())
	 {
	 	unsigned char** ubIn 	= ((RasterImage_gray*)imgIn)->getGray2D();
		unsigned char** ubOut 	= ((RasterImage_gray*)imgOut)->getGray2D();
		
		// just use imgIn limits since both are the same
		for (int nYi = nInYo; nYi <= nInYn; nYi++)
		{
			for (int nXi = nInXo; nXi <= nInXn; nXi++)
			{
				ubOut[nYi][nXi] = (unsigned char)(ImageStats.fMax - (ubIn[nYi][nXi] - ImageStats.fMin));
			}
		}	
	 }
	 else
	 {
	 	// mismatched types, FAIL for now
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same typed rasters accepted in copy(In,Out)");
     }

     // now set the output valid rectangle
     imgOut->setValidRect(imgIn->getValidRect());
	 	
} // end ImageMath::copy(RasterImage *imgIn, RasterImage *imgOut)

//-------------------------------

void ImageMath::transpose(RasterImage *imgIn)
{
	ImageRect *inRect  = imgIn->getValidRect();
	
	int nInXo  = inRect->getLeft(), 
		nInXn  = inRect->getRight(), 
	    nInYo  = inRect->getTop(), 
		nInYn  = inRect->getBottom();
		
	int nYi, nXi;	
		
		
	// make sure valid rect is square
    FAIL_CONDITION(inRect->getHeight() != inRect->getHeight(),
                kRasterImageError,
                "Input Valid Rect not square in transpose(In)");
	
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgIn->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input rasters accepted in transpose(In)");
                
	// for floats
	 if (imgIn->hasFloatRaster())
	 {
	 	float** fIn 	= ((RasterImage_gray_F*)imgIn)->getGrayF2D();
		float fTemp;
		
		for (nYi = nInYo; nYi <= nInYn; nYi++)
		{
			for (nXi = nInXo; nXi <= nYi; nXi++)
			{
				fTemp = fIn[nYi][nXi];
				fIn[nYi][nXi] = fIn[nXi][nYi];
				fIn[nXi][nYi] = fTemp;
			}
		}	
	 }
	 // for both unsigned chars
	 else if (!imgIn->hasFloatRaster())
	 {
	 	unsigned char** ubIn 	= ((RasterImage_gray*)imgIn)->getGray2D();
		unsigned char ubTemp;

		for (nYi = nInYo; nYi <= nInYn; nYi++)
		{
			for (nXi = nInXo; nXi <= nYi; nXi++)
			{
				ubTemp = ubIn[nYi][nXi];
				ubIn[nYi][nXi] = ubIn[nXi][nYi];
				ubIn[nXi][nYi] = ubTemp;
			}
		}	
	 }
	 else
	 {
	 	// mismatched types, FAIL for now
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same typed rasters accepted in copy(In,Out)");
     }
	
} // end ImageMath::transpose(RasterImage *imgIn)



//////////////////////////////////////////////////////////////////////////////////
//
// CORRELATE function
//
// filter size (B) MUST be odd in both dimensions
// filter size (B) must be smaller than A in both dimensions
//
void ImageMath::correlate(RasterImage *imgA, RasterImage *imgB, RasterImage *imgOut)
{

	ImageRect *pValidRectA   = imgA->getValidRect(),
		      *pValidRectB   = imgB->getValidRect(),
			  *pValidRectOut = imgOut->getValidRect();

	int nAXo   = pValidRectA->getLeft(),
		nAXn   = pValidRectA->getRight(),
		nAYo   = pValidRectA->getTop(),
		nAYn   = pValidRectA->getBottom();
	int nBXo   = pValidRectB->getLeft(),
		nBXn   = pValidRectB->getRight(),
		nBYo   = pValidRectB->getTop(),
		nBYn   = pValidRectB->getBottom();
	int nOutXo = pValidRectOut->getLeft(),
		nOutXn = pValidRectOut->getRight(),
		nOutYo = pValidRectOut->getTop(),
		nOutYn = pValidRectOut->getBottom();

	// make sure filter B sizes are odd, get radii in W and H
	short	wWradius, wFilterWidth = nBXn-nBXo+1;
	short	wHradius, wFilterHeight = nBYn-nBYo+1;
	FAIL_CONDITION(!kWLC_IsOdd(wFilterWidth),
                kRasterImageError,
                "Only odd width filters accepted in correlate(A,B,Out)");
	FAIL_CONDITION(!kWLC_IsOdd(wFilterHeight),
                kRasterImageError,
                "Only odd height filters accepted in correlate(A,B,Out)");
	wWradius = (wFilterWidth-1)/2;
	wHradius = (wFilterHeight-1)/2;

     // make sure input image (A) is larger than "filter" (B)           
    FAIL_CONDITION(imgA->getHeight() < imgB->getHeight(),
                kRasterImageError,
                "Input image has less rows than filter in correlate(A,B,Out)");
  	FAIL_CONDITION(imgA->getWidth() < imgB->getWidth(),
                kRasterImageError,
                "Input image has less columns than filter in correlate(A,B,Out)");

    // make sure output image is at least as large as input A
    // just check A vs Out since B must be smaller for convolve to make any sense           
    FAIL_CONDITION(imgOut->getHeight() < imgA->getHeight(),
                kRasterImageError,
                "Output image has less rows than inputs in correlate(A,B,Out)");
    FAIL_CONDITION(imgOut->getWidth() < imgA->getWidth(),
                kRasterImageError,
                "Output image has less columns than inputs in correlate(A,B,Out)");
	
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgA->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input A raster accepted in correlate(A,B,Out)");
	FAIL_CONDITION(imgB->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input B raster accepted in correlate(A,B,Out)");
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in correlate(A,B,Out)");

	// OK, it seems safe to try the filter now
	
	// get limits of operation based on valid image (A) size and filter (B) size
	int nXstart = nAXo+wWradius;
	int nXend   = nAXn-wWradius;
	int nYstart = nAYo+wHradius;
	int nYend   = nAYn-wHradius;
	float fBsqr = 0.0f;
	float fAsqr = 0.0f;
	float fDenom = 0.0f;
                
	// for all floats
	 if (imgA->hasFloatRaster() && imgB->hasFloatRaster() && imgOut->hasFloatRaster())
	 {
	 	float** fA 		= ((RasterImage_gray_F*)imgA)->getGrayF2D();
	 	float** fB	 	= ((RasterImage_gray_F*)imgB)->getGrayF2D();
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		int nYi, nXi, nYoff, nXoff, nYf, nXf;
		
		// compute the B sum of squares one time only
		for (nYf=nBYo; nYf<=nBYn; nYf++)
		{
			for (nXf=nBXo; nXf<=nBXn; nXf++)
			{
				fBsqr += fB[nYf][nXf]*fB[nYf][nXf];
			}
		}

	
		for (nYi = nYstart; nYi <= nYend; nYi++)
		{
			for (nXi = nXstart; nXi <= nXend; nXi++)
			{
				fOut[nYi][nXi] = 0.0f;
				nYoff = -wHradius;
				for (nYf=nBYo; nYf<=nBYn; nYf++)
				{
					nXoff = -wWradius;
					for (nXf=nBXo; nXf<=nBXn; nXf++)
					{
						fOut[nYi][nXi] += fA[nYi+nYoff][nXi+nXoff]*fB[nYf][nXf];
						fAsqr += fA[nYi+nYoff][nXi+nXoff]*fA[nYi+nYoff][nXi+nXoff];
						nXoff++;
					}
					nYoff++;
				}

				// now that sum of all products is done for this x&y, normalize
				fDenom = sqrt(fAsqr)*sqrt(fBsqr); 
				if (fDenom!=0) // we can normalize, if denom is zero, we're all messed up anyway
				{
					fOut[nYi][nXi] /= fDenom;
				}
				fAsqr = 0.0f; // reset here to allow for a debug break point
			}
		}	
	 }
	 // for all unsigned chars
	 else if (!imgA->hasFloatRaster() && !imgB->hasFloatRaster() && !imgOut->hasFloatRaster())
	 {
	 	unsigned char** ubA 	= ((RasterImage_gray*)imgA)->getGray2D();
	 	unsigned char** ubB	 	= ((RasterImage_gray*)imgB)->getGray2D();
		unsigned char** ubOut 	= ((RasterImage_gray*)imgOut)->getGray2D();
		int nYi, nXi, nYoff, nXoff, nYf, nXf;
		float fTemp;

		// compute the B sum of squares one time only
		for (nYf=0; nYf<wFilterHeight; nYf++)
		{
			for (nXf=0; nXf<wFilterWidth; nXf++)
			{
				fBsqr += ((float)ubB[nYf][nXf])*((float)ubB[nYf][nXf]);
			}
		}

		for (nYi = nYstart; nYi <= nYend; nYi++)
		{
			for (nXi = nXstart; nXi <= nXend; nXi++)
			{
				fTemp = 0.0f;
				nYoff = -wHradius;
				for (nYf=nBYo; nYf<=nBYn; nYf++)
				{
					nXoff = -wWradius;
					for (nXf=nBXo; nXf<=nBXn; nXf++)
					{
						fTemp += ((float)ubA[nYi+nYoff][nXi+nXoff])*((float)ubB[nYf][nXf]);
						fAsqr += ((float)ubA[nYi+nYoff][nXi+nXoff])*((float)ubA[nYi+nYoff][nXi+nXoff]);
						nXoff++;
					}
					nYoff++;
				}
				// now that sum of all products is done for this x&y, normalize
				fDenom = sqrt(fAsqr)*sqrt(fBsqr); 
				if (fDenom!=0) // we can normalize, if denom is zero, we're all messed up anyway
				{
					ubOut[nYi][nXi] = (unsigned char) fTemp/fDenom; 
				}
				else
				{
					ubOut[nYi][nXi] = (unsigned char) fTemp; 
				}
				fAsqr = 0.0f; // reset here to allow for a debug break point
				
			}
		}	
	 }
	 else
	 {
	 	// mismatched types, FAIL for now
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same typed rasters accepted in convolve(A,B,Out)");
     }
     
     // now set the output valid rectangle
     imgOut->setValidRect(nXstart, nYstart, nXend-nXstart+1, nYend-nYstart+1);
	 	
} // end ImageMath::correlate(RasterImage *imgA, RasterImage *imgB, RasterImage *imgOut)



//////////////////////////////////////////////////////////////////////////////////
//
// CONVOLVE function
//
// filter size (B) MUST be odd in both dimensions
// filter size (B) must be smaller than A in both dimensions
//
void ImageMath::convolve(RasterImage *imgA, RasterImage *imgB, RasterImage *imgOut)
{
	ImageRect *pValidRectA   = imgA->getValidRect(),
		      *pValidRectB   = imgB->getValidRect(),
			  *pValidRectOut = imgOut->getValidRect();

	int nAXo   = pValidRectA->getLeft(),
		nAXn   = pValidRectA->getRight(),
		nAYo   = pValidRectA->getTop(),
		nAYn   = pValidRectA->getBottom();
	int nBXo   = pValidRectB->getLeft(),
		nBXn   = pValidRectB->getRight(),
		nBYo   = pValidRectB->getTop(),
		nBYn   = pValidRectB->getBottom();
	int nOutXo = pValidRectOut->getLeft(),
		nOutXn = pValidRectOut->getRight(),
		nOutYo = pValidRectOut->getTop(),
		nOutYn = pValidRectOut->getBottom();

	// make sure filter B sizes are odd, get radii in W and H
	short	wWradius, wFilterWidth = nBXn-nBXo+1;
	short	wHradius, wFilterHeight = nBYn-nBYo+1;
	FAIL_CONDITION(!kWLC_IsOdd(wFilterWidth),
                kRasterImageError,
                "Only odd width filters accepted in convolve(A,B,Out)");
	FAIL_CONDITION(!kWLC_IsOdd(wFilterHeight),
                kRasterImageError,
                "Only odd height filters accepted in convolve(A,B,Out)");
	wWradius = (wFilterWidth-1)/2;
	wHradius = (wFilterHeight-1)/2;

     // make sure input image (A) is larger than "filter" (B)           
    FAIL_CONDITION(imgA->getHeight() < imgB->getHeight(),
                kRasterImageError,
                "Input image has less rows than filter in convolve(A,B,Out)");
  	FAIL_CONDITION(imgA->getWidth() < imgB->getWidth(),
                kRasterImageError,
                "Input image has less columns than filter in convolve(A,B,Out)");

    // make sure output image is at least as large as input A
    // just check A vs Out since B must be smaller for convolve to make any sense           
    FAIL_CONDITION(imgOut->getHeight() < imgA->getHeight(),
                kRasterImageError,
                "Output image has less rows than inputs in convolve(A,B,Out)");
    FAIL_CONDITION(imgOut->getWidth() < imgA->getWidth(),
                kRasterImageError,
                "Output image has less columns than inputs in convolve(A,B,Out)");
	
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgA->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input A raster accepted in convolve(A,B,Out)");
	FAIL_CONDITION(imgB->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input B raster accepted in convolve(A,B,Out)");
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in convolve(A,B,Out)");

	// OK, it seems safe to try the filter now
	
	// get limits of operation based on valid image (A) size and filter (B) size
	int nXstart = nAXo+wWradius;
	int nXend   = nAXn-wWradius;
	int nYstart = nAYo+wHradius;
	int nYend   = nAYn-wHradius;
                
	// for all floats
	 if (imgA->hasFloatRaster() && imgB->hasFloatRaster() && imgOut->hasFloatRaster())
	 {
	 	float** fA 		= ((RasterImage_gray_F*)imgA)->getGrayF2D();
	 	float** fB	 	= ((RasterImage_gray_F*)imgB)->getGrayF2D();
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		int nYi, nXi, nYoff, nXoff, nYf, nXf;
		
		for (nYi = nYstart; nYi <= nYend; nYi++)
		{
			for (nXi = nXstart; nXi <= nXend; nXi++)
			{
				fOut[nYi][nXi] = 0.0f;
				nYoff = -wHradius;
				for (nYf=nBYo; nYf<=nBYn; nYf++)
				{
					nXoff = -wWradius;
					for (nXf=nBXo; nXf<=nBXn; nXf++)
					{
						fOut[nYi][nXi] += fA[nYi+nYoff][nXi+nXoff]*fB[nYf][nXf];
						nXoff++;
					}
					nYoff++;
				}
			}
		}	
	 }
	 // for all unsigned chars
	 else if (!imgA->hasFloatRaster() && !imgB->hasFloatRaster() && !imgOut->hasFloatRaster())
	 {
	 	unsigned char** ubA 	= ((RasterImage_gray*)imgA)->getGray2D();
	 	unsigned char** ubB	 	= ((RasterImage_gray*)imgB)->getGray2D();
		unsigned char** ubOut 	= ((RasterImage_gray*)imgOut)->getGray2D();
		int nYi, nXi, nYf, nXf;

		for (nYi = nYstart; nYi <= nYend; nYi++)
		{
			for (nXi = nXstart; nXi <= nXend; nXi++)
			{
				ubOut[nYi][nXi] = 0.0f;
				int nYoff = -wHradius;
				for (nYf=nBYo; nYf<=nBYn; nYf++)
				{
					int nXoff = -wWradius;
					for (nXf=nBXo; nXf<=nBXn; nXf++)
					{
						ubOut[nYi][nXi] += ubA[nYi+nYoff][nXi+nXoff]*ubB[nYf][nXf];
						nXoff++;
					}
					nYoff++;
				}
			}
		}	
	 }
	 else
	 {
	 	// mismatched types, FAIL for now
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same typed rasters accepted in convolve(A,B,Out)");
     }
     
     // now set the output valid rectangle
     imgOut->setValidRect(nXstart, nYstart, nXend-nXstart+1, nYend-nYstart+1);
	 	
} // end ImageMath::convolve(RasterImage *imgA, RasterImage *imgB, RasterImage *imgOut)


//////////////////////////////////////////////////////////////////////////////////
//
// MULTIPLY functions
//
// (A) and (B) MUST be same size
// (O) must be >= (A) and (B) in both dimensions
//
void ImageMath::multiply(RasterImage *imgA, RasterImage *imgB, RasterImage *imgOut)
{
	ImageRect *pValidRectA   = imgA->getValidRect(),
		      *pValidRectB   = imgB->getValidRect(),
			  *pValidRectOut = imgOut->getValidRect();

	int nAXo   = pValidRectA->getLeft(),
		nAXn   = pValidRectA->getRight(),
		nAYo   = pValidRectA->getTop(),
		nAYn   = pValidRectA->getBottom();
	int nBXo   = pValidRectB->getLeft(),
		nBXn   = pValidRectB->getRight(),
		nBYo   = pValidRectB->getTop(),
		nBYn   = pValidRectB->getBottom();
	int nOutXo = pValidRectOut->getLeft(),
		nOutXn = pValidRectOut->getRight(),
		nOutYo = pValidRectOut->getTop(),
		nOutYn = pValidRectOut->getBottom();

     // make sure input image (A) is same size as (B)           
    FAIL_CONDITION(imgA->getHeight() != imgB->getHeight(),
                kRasterImageError,
                "Input images not same size in multiply(A,B,Out)");
  	FAIL_CONDITION(imgA->getWidth() != imgB->getWidth(),
                kRasterImageError,
                "Input images not same size in multiply(A,B,Out)");

    // make sure output image is at least as large as input A
    // just check A vs Out since A and B must be same size from above           
    FAIL_CONDITION(imgOut->getHeight() < imgA->getHeight(),
                kRasterImageError,
                "Output image has less rows than inputs in multiply(A,B,Out)");
    FAIL_CONDITION(imgOut->getWidth() < imgA->getWidth(),
                kRasterImageError,
                "Output image has less columns than inputs in multiply(A,B,Out)");
	
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgA->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input A raster accepted in multiply(A,B,Out)");
	FAIL_CONDITION(imgB->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input B raster accepted in multiply(A,B,Out)");
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in multiply(A,B,Out)");

	// OK, it seems safe to try the filter now
	
	// get limits of operation based on valid image (A) size
	int nXstart = nAXo;
	int nXend   = nAXn;
	int nYstart = nAYo;
	int nYend   = nAYn;
                
	// for all floats
	 if (imgA->hasFloatRaster() && imgB->hasFloatRaster() && imgOut->hasFloatRaster())
	 {
	 	float** fA 		= ((RasterImage_gray_F*)imgA)->getGrayF2D();
	 	float** fB	 	= ((RasterImage_gray_F*)imgB)->getGrayF2D();
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		int nYi, nXi;
		
		for (nYi = nYstart; nYi <= nYend; nYi++)
		{
			for (nXi = nXstart; nXi <= nXend; nXi++)
			{
				fOut[nYi][nXi] = fA[nYi][nXi]*fB[nYi][nXi];
			}
		}	
	 }
	 // for all unsigned chars
	 else if (!imgA->hasFloatRaster() && !imgB->hasFloatRaster() && !imgOut->hasFloatRaster())
	 {
	 	unsigned char** ubA 	= ((RasterImage_gray*)imgA)->getGray2D();
	 	unsigned char** ubB	 	= ((RasterImage_gray*)imgB)->getGray2D();
		unsigned char** ubOut 	= ((RasterImage_gray*)imgOut)->getGray2D();
		unsigned int    unTemp;
		
		for (int nYi = nYstart; nYi <= nYend; nYi++)
		{
			for (int nXi = nXstart; nXi <= nXend; nXi++)
			{
				unTemp = ubA[nYi][nXi]*ubB[nYi][nXi];
				ubOut[nYi][nXi] = kWLC_Min(unTemp,255);
			}
		}	
	 }
	 else
	 {
	 	// mismatched types, FAIL for now
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same typed rasters accepted in multiply(A,B,Out)");
     }
     
     // now set the output valid rectangle
     imgOut->setValidRect(nXstart, nYstart, nXend-nXstart+1, nYend-nYstart+1);
	 	
} // end ImageMath::multiply(RasterImage *imgA, RasterImage *imgB, RasterImage *imgOut)


//-------------------------------


void ImageMath::multiplyComplex(RasterImage *imgA, RasterImage *imgB, RasterImage *imgOut)
{
	ImageRect *pValidRectA   = imgA->getValidRect(),
		      *pValidRectB   = imgB->getValidRect(),
			  *pValidRectOut = imgOut->getValidRect();

	int nAXo   = pValidRectA->getLeft(),
		nAXn   = pValidRectA->getRight(),
		nAYo   = pValidRectA->getTop(),
		nAYn   = pValidRectA->getBottom();
	int nBXo   = pValidRectB->getLeft(),
		nBXn   = pValidRectB->getRight(),
		nBYo   = pValidRectB->getTop(),
		nBYn   = pValidRectB->getBottom();
	int nOutXo = pValidRectOut->getLeft(),
		nOutXn = pValidRectOut->getRight(),
		nOutYo = pValidRectOut->getTop(),
		nOutYn = pValidRectOut->getBottom();

     // make sure input image (A) is same size as (B)           
    FAIL_CONDITION(imgA->getHeight() != imgB->getHeight(),
                kRasterImageError,
                "Input images not same size in multiply(A,B,Out)");
  	FAIL_CONDITION(imgA->getWidth() != imgB->getWidth(),
                kRasterImageError,
                "Input images not same size in multiply(A,B,Out)");

    // make sure output image is at least as large as input A
    // just check A vs Out since A and B must be same size from above           
    FAIL_CONDITION(imgOut->getHeight() < imgA->getHeight(),
                kRasterImageError,
                "Output image has less rows than inputs in multiply(A,B,Out)");
    FAIL_CONDITION(imgOut->getWidth() < imgA->getWidth(),
                kRasterImageError,
                "Output image has less columns than inputs in multiply(A,B,Out)");
	
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgA->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input A raster accepted in multiply(A,B,Out)");
	FAIL_CONDITION(imgB->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input B raster accepted in multiply(A,B,Out)");
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in multiply(A,B,Out)");

	// OK, it seems safe to try the filter now
	
	// get limits of operation based on valid image (A) size
	int nXstart = nAXo;
	int nXend   = nAXn;
	int nYstart = nAYo;
	int nYend   = nAYn;
                
	// for all floats
	 if (imgA->hasFloatRaster() && imgB->hasFloatRaster() && imgOut->hasFloatRaster())
	 {
	 	float** fA 		= ((RasterImage_gray_F*)imgA)->getGrayF2D();
	 	float** fB	 	= ((RasterImage_gray_F*)imgB)->getGrayF2D();
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		int nYi, nXi;
		
		for (nYi = nYstart; nYi <= nYend; nYi++)
		{
			for (nXi = nXstart; nXi <= nXend; nXi+=2)
			{
				fOut[nYi][nXi]   = fA[nYi][nXi]  *fB[nYi][nXi] - fA[nYi][nXi+1]*fB[nYi][nXi+1];
				fOut[nYi][nXi+1] = fA[nYi][nXi+1]*fB[nYi][nXi] + fA[nYi][nXi]  *fB[nYi][nXi+1];
			}
		}	
	 }
	 // for all unsigned chars
	 else if (!imgA->hasFloatRaster() && !imgB->hasFloatRaster() && !imgOut->hasFloatRaster())
	 {
	 	unsigned char** ubA 	= ((RasterImage_gray*)imgA)->getGray2D();
	 	unsigned char** ubB	 	= ((RasterImage_gray*)imgB)->getGray2D();
		unsigned char** ubOut 	= ((RasterImage_gray*)imgOut)->getGray2D();
		unsigned int    unTemp;
		
		for (int nYi = nYstart; nYi <= nYend; nYi++)
		{
			for (int nXi = nXstart; nXi <= nXend; nXi+=2)
			{
				unTemp = ubA[nYi][nXi]*ubB[nYi][nXi] - ubA[nYi][nXi+1]*ubB[nYi][nXi+1];
				ubOut[nYi][nXi] = kWLC_Min(unTemp,255);
				unTemp = ubA[nYi][nXi+1]*ubB[nYi][nXi] + ubA[nYi][nXi]*ubB[nYi][nXi+1];
				ubOut[nYi][nXi+1] = kWLC_Min(unTemp,255);
			}
		}	
	 }
	 else
	 {
	 	// mismatched types, FAIL for now
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same typed rasters accepted in multiply(A,B,Out)");
     }
     
     // now set the output valid rectangle
     imgOut->setValidRect(nXstart, nYstart, nXend-nXstart+1, nYend-nYstart+1);
	 	
} // end ImageMath::multiply(RasterImage *imgA, RasterImage *imgB, RasterImage *imgOut)


//////////////////////////////////////////////////////////////////////////////////
//
// SCALE functions
//
// (O) must be >= (A) in both dimensions
//
void ImageMath::scale(RasterImage *imgA, float fScale)
{
	ImageRect *pValidRectA   = imgA->getValidRect();
	int nAXo   = pValidRectA->getLeft(),
		nAXn   = pValidRectA->getRight(),
		nAYo   = pValidRectA->getTop(),
		nAYn   = pValidRectA->getBottom();

	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgA->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input A raster accepted in scale(A,fScale,Out)");

	// OK, it seems safe to try the filter now
	
	// get limits of operation based on valid image (A) size
	int nXstart = nAXo;
	int nXend   = nAXn;
	int nYstart = nAYo;
	int nYend   = nAYn;
                
	// for all floats
	 if (imgA->hasFloatRaster())
	 {
	 	float** fA 		= ((RasterImage_gray_F*)imgA)->getGrayF2D();
		int nYi, nXi;
		
		for (nYi = nYstart; nYi <= nYend; nYi++)
		{
			for (nXi = nXstart; nXi <= nXend; nXi++)
			{
				fA[nYi][nXi] *= fScale;
			}
		}	
	 }
	 // for all unsigned chars
	 else if (!imgA->hasFloatRaster())
	 {
	 	unsigned char** ubA 	= ((RasterImage_gray*)imgA)->getGray2D();
		unsigned int    unTemp;
		
		for (int nYi = nYstart; nYi <= nYend; nYi++)
		{
			for (int nXi = nXstart; nXi <= nXend; nXi++)
			{
				unTemp = (unsigned int)(ubA[nYi][nXi]*fScale);
				ubA[nYi][nXi] = kWLC_Max(0,kWLC_Min(unTemp,255));
			}
		}	
	 }
	 else
	 {
	 	// mismatched types, FAIL for now
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same typed rasters accepted in scale(A,fScale,Out)");
     }
     
} // end ImageMath::scale(RasterImage *imgA, float fScale)


//-------------------------------


void ImageMath::scale(RasterImage *imgA, float fScale, RasterImage *imgOut)
{
	ImageRect *pValidRectA   = imgA->getValidRect(),
			  *pValidRectOut = imgOut->getValidRect();

	int nAXo   = pValidRectA->getLeft(),
		nAXn   = pValidRectA->getRight(),
		nAYo   = pValidRectA->getTop(),
		nAYn   = pValidRectA->getBottom();
	int nOutXo = pValidRectOut->getLeft(),
		nOutXn = pValidRectOut->getRight(),
		nOutYo = pValidRectOut->getTop(),
		nOutYn = pValidRectOut->getBottom();

    // make sure output image is at least as large as input A
    FAIL_CONDITION(imgOut->getHeight() < imgA->getHeight(),
                kRasterImageError,
                "Output image has less rows than inputs in scale(A,fScale,Out)");
    FAIL_CONDITION(imgOut->getWidth() < imgA->getWidth(),
                kRasterImageError,
                "Output image has less columns than inputs in scale(A,fScale,Out)");
	
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgA->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input A raster accepted in scale(A,fScale,Out)");
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in scale(A,fScale,Out)");

	// OK, it seems safe to try the filter now
	
	// get limits of operation based on valid image (A) size
	int nXstart = nAXo;
	int nXend   = nAXn;
	int nYstart = nAYo;
	int nYend   = nAYn;
                
	// for all floats
	 if (imgA->hasFloatRaster() && imgOut->hasFloatRaster())
	 {
	 	float** fA 		= ((RasterImage_gray_F*)imgA)->getGrayF2D();
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		int nYi, nXi;
		
		for (nYi = nYstart; nYi <= nYend; nYi++)
		{
			for (nXi = nXstart; nXi <= nXend; nXi++)
			{
				fOut[nYi][nXi] = fA[nYi][nXi]*fScale;
			}
		}	
	 }
	 // for all unsigned chars
	 else if (!imgA->hasFloatRaster() && !imgOut->hasFloatRaster())
	 {
	 	unsigned char** ubA 	= ((RasterImage_gray*)imgA)->getGray2D();
		unsigned char** ubOut 	= ((RasterImage_gray*)imgOut)->getGray2D();
		unsigned int    unTemp;
		
		for (int nYi = nYstart; nYi <= nYend; nYi++)
		{
			for (int nXi = nXstart; nXi <= nXend; nXi++)
			{
				unTemp = (unsigned int)(ubA[nYi][nXi]*fScale);
				ubOut[nYi][nXi] = kWLC_Max(0,kWLC_Min(unTemp,255));
			}
		}	
	 }
	 else
	 {
	 	// mismatched types, FAIL for now
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same typed rasters accepted in scale(A,fScale,Out)");
     }
     
     // now set the output valid rectangle
     imgOut->setValidRect(nXstart, nYstart, nXend-nXstart+1, nYend-nYstart+1);
	 	
} // end ImageMath::scale(RasterImage *imgA, float fScale, RasterImage *imgOut)

//////////////////////////////////////////////////////////////////////////////////
//
// ADD functions

void ImageMath::addInto(RasterImage *imgIn, RasterImage *imgOut)
{
	ImageRect *inRect  = imgIn->getValidRect(),
			  *outRect = imgOut->getValidRect();

	int nInXo  = inRect->getLeft(), 
		nOutXo = outRect->getLeft(),
		nInXn  = inRect->getRight(), 
		nOutXn = outRect->getRight();
	int nInYo  = inRect->getTop(), 
		nOutYo = outRect->getTop(),
		nInYn  = inRect->getBottom(), 
		nOutYn = outRect->getBottom();
		
	// make sure valid rects are the same
	if ( (nInXo!=nOutXo) || (nInXn!=nOutXn) || (nInYo!=nOutYo) || (nInYn!=nOutYn) )
	{
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same size valid rect rasters accepted in addInto(In,Out)");
	}		
	
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgIn->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input rasters accepted in addInto(In,Out)");
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in addInto(In,Out)");
                
	// for both floats
	 if (imgIn->hasFloatRaster() && imgOut->hasFloatRaster())
	 {
	 	float** fIn 	= ((RasterImage_gray_F*)imgIn)->getGrayF2D();
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		
		// just use imgIn limits since both are the same
		for (int nYi = nInYo; nYi <= nInYn; nYi++)
		{
			for (int nXi = nInXo; nXi <= nInXn; nXi++)
			{
				fOut[nYi][nXi] += fIn[nYi][nXi];
			}
		}	
	 }
	 // for both unsigned chars
	 else if (!imgIn->hasFloatRaster() && !imgOut->hasFloatRaster())
	 {
	 	unsigned char** ubIn 	= ((RasterImage_gray*)imgIn)->getGray2D();
		unsigned char** ubOut 	= ((RasterImage_gray*)imgOut)->getGray2D();
		
		// just use imgIn limits since both are the same
		for (int nYi = nInYo; nYi <= nInYn; nYi++)
		{
			for (int nXi = nInXo; nXi <= nInXn; nXi++)
			{
				ubOut[nYi][nXi] += ubIn[nYi][nXi];
			}
		}	
	 }
	 else
	 {
	 	// mismatched types, FAIL for now
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same typed rasters accepted in addInto(In,Out)");
     }

     // now set the output valid rectangle
     imgOut->setValidRect(imgIn->getValidRect());
	 	
} // end ImageMath::addInto(RasterImage *imgIn, RasterImage *imgOut)


//-------------------------------

void ImageMath::add(RasterImage *imgA, RasterImage *imgB, RasterImage *imgOut)
{
	ImageRect *aRect = imgA->getValidRect(),
			  *bRect = imgB->getValidRect();

	int nAXo = aRect->getLeft(), 
		nBXo = bRect->getLeft(),
		nAXn = aRect->getRight(), 
		nBXn = bRect->getRight();
	int nAYo = aRect->getTop(), 
		nBYo = bRect->getTop(),
		nAYn = aRect->getBottom(), 
		nBYn = bRect->getBottom();
		
	// make sure valid rects are the same
	if ( (nAXo!=nBXo) || (nAXn!=nBXn) || (nAYo!=nBYo) || (nAYn!=nBYn) )
	{
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same size valid rect rasters accepted in add(A,B,Out)");
	}		
	
    // make sure output image is at least as large as inputs
    // just check A vs Out since A & B were checked for valid rects            
    FAIL_CONDITION(imgOut->getHeight() < imgA->getHeight(),
                kRasterImageError,
                "Output image has less rows than inputs in add(A,B,Out)");
    FAIL_CONDITION(imgOut->getWidth() < imgA->getWidth(),
                kRasterImageError,
                "Output image has less columns than inputs in add(A,B,Out)");
	
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgA->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input A raster accepted in add(A,B,Out)");
	FAIL_CONDITION(imgB->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input B raster accepted in add(A,B,Out)");
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in add(A,B,Out)");
                
	// for all floats
	 if (imgA->hasFloatRaster() && imgB->hasFloatRaster() && imgOut->hasFloatRaster())
	 {
	 	float** fA 		= ((RasterImage_gray_F*)imgA)->getGrayF2D();
	 	float** fB	 	= ((RasterImage_gray_F*)imgB)->getGrayF2D();
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		
		// just use imgA limits since both are the same
		for (int nYi = nAYo; nYi <= nAYn; nYi++)
		{
			for (int nXi = nAXo; nXi <= nAXn; nXi++)
			{
				fOut[nYi][nXi] = fA[nYi][nXi] + fB[nYi][nXi];
			}
		}	
	 }
	 // for all unsigned chars
	 else if (!imgA->hasFloatRaster() && !imgB->hasFloatRaster() && !imgOut->hasFloatRaster())
	 {
	 	unsigned char** ubA 	= ((RasterImage_gray*)imgA)->getGray2D();
	 	unsigned char** ubB	 	= ((RasterImage_gray*)imgB)->getGray2D();
		unsigned char** ubOut 	= ((RasterImage_gray*)imgOut)->getGray2D();
		
		// just use imgA limits since both are the same
		for (int nYi = nAYo; nYi <= nAYn; nYi++)
		{
			for (int nXi = nAXo; nXi <= nAXn; nXi++)
			{
				ubOut[nYi][nXi] = ubA[nYi][nXi] + ubB[nYi][nXi];
			}
		}	
	 }
	 else
	 {
	 	// mismatched types, FAIL for now
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same typed rasters accepted in add(A,B,Out)");
     }

     // now set the output valid rectangle
     imgOut->setValidRect(imgA->getValidRect());
	 	
} // end ImageMath::addInto(RasterImage *imgA, RasterImage *imgB, RasterImage *imgOut)


//////////////////////////////////////////////////////////////////////////////////
//
// SUBTRACT functions

void ImageMath::subtractFrom(RasterImage *imgIn, RasterImage *imgOut)
{
	ImageRect *inRect  = imgIn->getValidRect(),
			  *outRect = imgOut->getValidRect();

	int nInXo  = inRect->getLeft(), 
		nOutXo = outRect->getLeft(),
		nInXn  = inRect->getRight(), 
		nOutXn = outRect->getRight();
	int nInYo  = inRect->getTop(), 
		nOutYo = outRect->getTop(),
		nInYn  = inRect->getBottom(), 
		nOutYn = outRect->getBottom();
		
	// make sure valid rects are the same
	if ( (nInXo!=nOutXo) || (nInXn!=nOutXn) || (nInYo!=nOutYo) || (nInYn!=nOutYn) )
	{
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same size valid rect rasters accepted in subtractFrom(In,Out)");
	}		
	
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgIn->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input rasters accepted in subtractFrom(In,Out)");
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in subtractFrom(In,Out)");
                
	// for both floats
	 if (imgIn->hasFloatRaster() && imgOut->hasFloatRaster())
	 {
	 	float** fIn 	= ((RasterImage_gray_F*)imgIn)->getGrayF2D();
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		
		// just use imgIn limits since both are the same
		for (int nYi = nInYo; nYi <= nInYn; nYi++)
		{
			for (int nXi = nInXo; nXi <= nInXn; nXi++)
			{
				fOut[nYi][nXi] -= fIn[nYi][nXi];
			}
		}	
	 }
	 // for both unsigned chars
	 else if (!imgIn->hasFloatRaster() && !imgOut->hasFloatRaster())
	 {
	 	unsigned char** ubIn 	= ((RasterImage_gray*)imgIn)->getGray2D();
		unsigned char** ubOut 	= ((RasterImage_gray*)imgOut)->getGray2D();
		
		// just use imgIn limits since both are the same
		for (int nYi = nInYo; nYi <= nInYn; nYi++)
		{
			for (int nXi = nInXo; nXi <= nInXn; nXi++)
			{
				ubOut[nYi][nXi] -= ubIn[nYi][nXi];
			}
		}	
	 }
	 else
	 {
	 	// mismatched types, FAIL for now
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same typed rasters accepted in subtractFrom(In,Out)");
     }

     // now set the output valid rectangle
     imgOut->setValidRect(imgIn->getValidRect());
	 	
} // end ImageMath::subtractFrom(RasterImage *imgIn, RasterImage *imgOut)

//-------------------------------

void ImageMath::subtract(RasterImage *imgA, RasterImage *imgB, RasterImage *imgOut)
{
	ImageRect *aRect = imgA->getValidRect(),
			  *bRect = imgB->getValidRect();

	int nAXo = aRect->getLeft(), 
		nBXo = bRect->getLeft(),
		nAXn = aRect->getRight(), 
		nBXn = bRect->getRight();
	int nAYo = aRect->getTop(), 
		nBYo = bRect->getTop(),
		nAYn = aRect->getBottom(), 
		nBYn = bRect->getBottom();
		
	// make sure valid rects are the same
	if ( (nAXo!=nBXo) || (nAXn!=nBXn) || (nAYo!=nBYo) || (nAYn!=nBYn) )
	{
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same size valid rect rasters accepted in subtract(A,B,Out)");
	}		
	
    // make sure output image is at least as large as inputs
    // just check A vs Out since A & B were checked for valid rects            
    FAIL_CONDITION(imgOut->getHeight() < imgA->getHeight(),
                kRasterImageError,
                "Output image has less rows than inputs in subtract(A,B,Out)");
    FAIL_CONDITION(imgOut->getWidth() < imgA->getWidth(),
                kRasterImageError,
                "Output image has less columns than inputs in subtract(A,B,Out)");
	
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgA->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input A raster accepted in subtract(A,B,Out)");
	FAIL_CONDITION(imgB->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input B raster accepted in subtract(A,B,Out)");
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in subtract(A,B,Out)");
                
	// for all floats
	 if (imgA->hasFloatRaster() && imgB->hasFloatRaster() && imgOut->hasFloatRaster())
	 {
	 	float** fA 		= ((RasterImage_gray_F*)imgA)->getGrayF2D();
	 	float** fB	 	= ((RasterImage_gray_F*)imgB)->getGrayF2D();
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		
		// just use imgA limits since both are the same
		for (int nYi = nAYo; nYi <= nAYn; nYi++)
		{
			for (int nXi = nAXo; nXi <= nAXn; nXi++)
			{
				fOut[nYi][nXi] = fA[nYi][nXi] - fB[nYi][nXi];
			}
		}	
	 }
	 // for all unsigned chars
	 else if (!imgA->hasFloatRaster() && !imgB->hasFloatRaster() && !imgOut->hasFloatRaster())
	 {
	 	unsigned char** ubA 	= ((RasterImage_gray*)imgA)->getGray2D();
	 	unsigned char** ubB	 	= ((RasterImage_gray*)imgB)->getGray2D();
		unsigned char** ubOut 	= ((RasterImage_gray*)imgOut)->getGray2D();
		
		// just use imgA limits since both are the same
		for (int nYi = nAYo; nYi <= nAYn; nYi++)
		{
			for (int nXi = nAXo; nXi <= nAXn; nXi++)
			{
				ubOut[nYi][nXi] = ubA[nYi][nXi] - ubB[nYi][nXi];
			}
		}	
	 }
	 else
	 {
	 	// mismatched types, FAIL for now
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same typed rasters accepted in subtract(A,B,Out)");
     }

     // now set the output valid rectangle
     imgOut->setValidRect(imgA->getValidRect());
	 	
} // end ImageMath::subtract(RasterImage *imgA, RasterImage *imgB, RasterImage *imgOut)


//////////////////////////////////////////////////////////////////////////////////
//
// MAX functions

void ImageMath::maxInto(RasterImage *imgIn, RasterImage *imgOut)
{
	ImageRect *inRect  = imgIn->getValidRect(),
			  *outRect = imgOut->getValidRect();

	int nInXo  = inRect->getLeft(), 
		nOutXo = outRect->getLeft(),
		nInXn  = inRect->getRight(), 
		nOutXn = outRect->getRight();
	int nInYo  = inRect->getTop(), 
		nOutYo = outRect->getTop(),
		nInYn  = inRect->getBottom(), 
		nOutYn = outRect->getBottom();
		
	// make sure valid rects are the same
	if ( (nInXo!=nOutXo) || (nInXn!=nOutXn) || (nInYo!=nOutYo) || (nInYn!=nOutYn) )
	{
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same size valid rect rasters accepted in maxInto(In,Out)");
	}		
	
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgIn->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input rasters accepted in maxInto(In,Out)");
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in maxInto(In,Out)");
                
	// for both floats
	 if (imgIn->hasFloatRaster() && imgOut->hasFloatRaster())
	 {
	 	float** fIn 	= ((RasterImage_gray_F*)imgIn)->getGrayF2D();
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		
		// just use imgIn limits since both are the same
		for (int nYi = nInYo; nYi <= nInYn; nYi++)
		{
			for (int nXi = nInXo; nXi <= nInXn; nXi++)
			{
				fOut[nYi][nXi] = kWLC_Max(fIn[nYi][nXi],fOut[nYi][nXi]);
			}
		}	
	 }
	 // for both unsigned chars
	 else if (!imgIn->hasFloatRaster() && !imgOut->hasFloatRaster())
	 {
	 	unsigned char** ubIn 	= ((RasterImage_gray*)imgIn)->getGray2D();
		unsigned char** ubOut 	= ((RasterImage_gray*)imgOut)->getGray2D();
		
		// just use imgIn limits since both are the same
		for (int nYi = nInYo; nYi <= nInYn; nYi++)
		{
			for (int nXi = nInXo; nXi <= nInXn; nXi++)
			{
				ubOut[nYi][nXi] = kWLC_Max(ubIn[nYi][nXi],ubOut[nYi][nXi]);
			}
		}	
	 }
	 else
	 {
	 	// mismatched types, FAIL for now
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same typed rasters accepted in maxInto(In,Out)");
     }

     // now set the output valid rectangle
     imgOut->setValidRect(imgIn->getValidRect());
	 	
} // end ImageMath::maxInto(RasterImage *imgIn, RasterImage *imgOut)

//-------------------------------

void ImageMath::maxOf(RasterImage *imgA, RasterImage *imgB, RasterImage *imgOut)
{
	ImageRect *aRect = imgA->getValidRect(),
			  *bRect = imgB->getValidRect();

	int nAXo = aRect->getLeft(), 
		nBXo = bRect->getLeft(),
		nAXn = aRect->getRight(), 
		nBXn = bRect->getRight();
	int nAYo = aRect->getTop(), 
		nBYo = bRect->getTop(),
		nAYn = aRect->getBottom(), 
		nBYn = bRect->getBottom();
		
	// make sure valid rects are the same
	if ( (nAXo!=nBXo) || (nAXn!=nBXn) || (nAYo!=nBYo) || (nAYn!=nBYn) )
	{
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same size valid rect rasters accepted in maxOf(A,B,Out)");
	}		
	
    // make sure output image is at least as large as inputs
    // just check A vs Out since A & B were checked for valid rects            
    FAIL_CONDITION(imgOut->getHeight() < imgA->getHeight(),
                kRasterImageError,
                "Output image has less rows than inputs in maxOf(A,B,Out)");
    FAIL_CONDITION(imgOut->getWidth() < imgA->getWidth(),
                kRasterImageError,
                "Output image has less columns than inputs in maxOf(A,B,Out)");
	
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgA->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input A raster accepted in maxOf(A,B,Out)");
	FAIL_CONDITION(imgB->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input B raster accepted in maxOf(A,B,Out)");
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in maxOf(A,B,Out)");
                
	// for all floats
	 if (imgA->hasFloatRaster() && imgB->hasFloatRaster() && imgOut->hasFloatRaster())
	 {
	 	float** fA 		= ((RasterImage_gray_F*)imgA)->getGrayF2D();
	 	float** fB	 	= ((RasterImage_gray_F*)imgB)->getGrayF2D();
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		
		// just use imgA limits since both are the same
		for (int nYi = nAYo; nYi <= nAYn; nYi++)
		{
			for (int nXi = nAXo; nXi <= nAXn; nXi++)
			{
				fOut[nYi][nXi] = kWLC_Max(fA[nYi][nXi],fB[nYi][nXi]);
			}
		}	
	 }
	 // for all unsigned chars
	 else if (!imgA->hasFloatRaster() && !imgB->hasFloatRaster() && !imgOut->hasFloatRaster())
	 {
	 	unsigned char** ubA 	= ((RasterImage_gray*)imgA)->getGray2D();
	 	unsigned char** ubB	 	= ((RasterImage_gray*)imgB)->getGray2D();
		unsigned char** ubOut 	= ((RasterImage_gray*)imgOut)->getGray2D();
		
		// just use imgA limits since both are the same
		for (int nYi = nAYo; nYi <= nAYn; nYi++)
		{
			for (int nXi = nAXo; nXi <= nAXn; nXi++)
			{
				ubOut[nYi][nXi] = kWLC_Max(ubA[nYi][nXi],ubB[nYi][nXi]);
			}
		}	
	 }
	 else
	 {
	 	// mismatched types, FAIL for now
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only same typed rasters accepted in maxOf(A,B,Out)");
     }

     // now set the output valid rectangle
     imgOut->setValidRect(imgA->getValidRect());
	 	
} // end ImageMath::maxOf(RasterImage *imgA, RasterImage *imgB, RasterImage *imgOut)

//////////////////////////////////////////////////////////////////////////////////
//
// COMBINE functions

void ImageMath::combine(RasterImage *imgA, int nScale, RasterImage *imgOut)
{
	ImageRect *aRect = imgA->getValidRect(),
			  *oRect = imgOut->getBoundRect();

	int nAXo = aRect->getLeft(),
		nAYo = aRect->getTop(), 
		nAXn = aRect->getRight(),
		nAYn = aRect->getBottom(),
		nAWidth = aRect->getWidth(),
		nAHeight = aRect->getHeight(); 
		 
	int nOXo = oRect->getLeft(),
		nOYo = oRect->getTop(), 
		nOXn = oRect->getRight(),
		nOYn = oRect->getBottom(), 
		nOWidth = oRect->getWidth(),
		nOHeight = oRect->getHeight();
		
	int nXO, nYO;
 
		
	// make sure output is big enough to hold combined input
	if ( (nOWidth < nAWidth/nScale) || (nOHeight < nAHeight/nScale) )
	{
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Output raster too small in combine(A,scale,Out)");
	}		
	
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgA->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input A raster accepted in combine(A,scale,Out)");
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in combine(A,scale,Out)");
                
	// for all floats
	 if (imgA->hasFloatRaster() && imgOut->hasFloatRaster())
	 {
	 	float** fA 		= ((RasterImage_gray_F*)imgA)->getGrayF2D();
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		
		nYO = 0;
		for (int nYi = nAYo; nYi <= nAYn; nYi+=nScale)
		{
			int nYlimit = nScale<=(nAHeight - nYO*nScale)?nScale:(nAHeight - nYO*nScale);
			nXO = 0;
			for (int nXi = nAXo; nXi <= nAXn; nXi+=nScale)
			{
				int nXlimit = nScale<=(nAWidth - nXO*nScale)?nScale:(nAWidth - nXO*nScale);
				fOut[nYO][nXO] = 0.0f;
				int nCount = 0;
				for (int iy=0; iy<nYlimit; iy++)
				{
					for (int jx=0; jx<nXlimit; jx++)
					{
						fOut[nYO][nXO] += fA[nYi+iy][nXi+jx];
						nCount ++;
					}
				} // end loop over combining items
				fOut[nYO][nXO] /= nCount; // get average
				nXO++; // advance to next output column
			}
			nYO++; // advance to next output row
			
		} // end loop over input raster
			
	 } // endif both are float rasters
	 else
	 {
	 	// mismatched types, FAIL for now
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only float rasters accepted in combine(A,scale,Out)");
     }

     // now set the output valid rectangle
     imgOut->setValidRect(0,0,nXO,nYO);
	 	
} // end ImageMath::combine(RasterImage *imgA, int nScale, RasterImage *imgOut)

//-------------------------------

void ImageMath::uncombine(RasterImage *imgA, int nScale, RasterImage *imgOut)
{
	ImageRect *aRect = imgA->getValidRect(),
			  *oRect = imgOut->getBoundRect();

	int nAXo = aRect->getLeft(),
		nAYo = aRect->getTop(), 
		nAXn = aRect->getRight(),
		nAYn = aRect->getBottom(),
		nAWidth = aRect->getWidth(),
		nAHeight = aRect->getHeight(); 
		 
	int nOXo = oRect->getLeft(),
		nOYo = oRect->getTop(), 
		nOXn = oRect->getRight(),
		nOYn = oRect->getBottom(), 
		nOWidth = oRect->getWidth(),
		nOHeight = oRect->getHeight();
		
	int nXO, nYO;
 
		
	// make sure output is big enough to hold combined input
	if ( (nOWidth < nAWidth*nScale) || (nOHeight < nAHeight*nScale) )
	{
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Output raster too small in uncombine(A,scale,Out)");
	}		
	
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgA->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input A raster accepted in uncombine(A,scale,Out)");
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in uncombine(A,scale,Out)");
                
	// for all floats
	 if (imgA->hasFloatRaster() && imgOut->hasFloatRaster())
	 {
	 	float** fA 		= ((RasterImage_gray_F*)imgA)->getGrayF2D();
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		
		for (int nYi = nAYo; nYi <= nAYn; nYi++)
		{
			for (int nXi = nAXo; nXi <= nAXn; nXi++)
			{
				nYO = nYi*nScale;
				for (int iy=0; iy<nScale; iy++)
				{
					nXO = nXi*nScale;
					for (int jx=0; jx<nScale; jx++)
					{
						fOut[nYO][nXO] = fA[nYi][nXi];
						nXO++;
					}
					nYO++;
				} // end loop over combining items
			}
		} // end loop over input raster
			
	 } // endif both are float rasters
	 else
	 {
	 	// mismatched types, FAIL for now
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only float rasters accepted in uncombine(A,scale,Out)");
     }

     // now set the output valid rectangle
     imgOut->setValidRect(0,0,(nAXn-nAXo+1)*nScale,(nAYn-nAXo+1)*nScale);
	 	
} // end ImageMath::uncombine(RasterImage *imgA, int nScale, RasterImage *imgOut)


//////////////////////////////////////////////////////////////////////////////////
//
// SELECT_REGION function

void ImageMath::selectRegion(RasterImage *imgA, int nXc, int nYc, RasterImage *imgOut)
{
	ImageRect *aRect = imgA->getValidRect(),
			  *oRect = imgOut->getValidRect();

	int nAXo = aRect->getLeft(),
		nAYo = aRect->getTop(), 
		nAXn = aRect->getRight(),
		nAYn = aRect->getBottom(),
		nAWidth = aRect->getWidth(),
		nAHeight = aRect->getHeight(); 
		 
	int nOXo = oRect->getLeft(),
		nOYo = oRect->getTop(), 
		nOXn = oRect->getRight(),
		nOYn = oRect->getBottom(), 
		nOWidth = oRect->getWidth(),
		nOHeight = oRect->getHeight();
		
	int nXO, nYO, nHalfWidth = nOWidth/2, nHalfHeight = nOHeight/2;
 
		
	// make sure input is at least as big as desired output
	if ( (nAWidth < nOWidth) || (nAHeight < nOHeight) )
	{
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Input raster too small in selectRegion(A,Xc,Yc,Out)");
	}
			
	
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgA->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level input A raster accepted in selectRegion(A,Xc,Yc,Out)");
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in selectRegion(A,Xc,Yc,Out)");
     
    // if nXc and/or nYc are too close to an input edge, just move them in some
    if (nXc<nHalfWidth) 		  nXc = nHalfWidth;
    if (nAWidth-nXc<nHalfWidth)   nXc = nAWidth-nHalfWidth;
    if (nYc<nHalfHeight)		  nYc = nHalfHeight;
    if (nAHeight-nYc<nHalfHeight) nYc = nAHeight-nHalfHeight;
                
	// for all floats
	 if (imgA->hasFloatRaster() && imgOut->hasFloatRaster())
	 {
	 	float** fA 		= ((RasterImage_gray_F*)imgA)->getGrayF2D();
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		
		nYO = 0;
		for (int nYi = nYc-nHalfHeight; nYi < nYc+nHalfHeight; nYi++)
		{
			nXO = 0;
			for (int nXi = nXc-nHalfWidth; nXi < nXc+nHalfWidth; nXi++)
			{
				fOut[nYO][nXO] = fA[nYi][nXi];
				nXO++; // advance to next output column
			}
			nYO++; // advance to next output row

		} // end loop over input raster
			
	 } // endif both are float rasters
	 else
	 {
	 	// mismatched types, FAIL for now
		FAIL_CONDITION(TRUE,
                kRasterImageError,
                "Only float rasters accepted in combine(A,scale,Out)");
     }

     // now set the output valid rectangle
     imgOut->setValidRect(0,0,nXO,nYO);
	 	
} // end ImageMath::selectRegion(RasterImage *imgA, int nXc, int nYc, RasterImage *imgOut)

//////////////////////////////////////////////////////////////////////////////////
//
// FILL functions

void ImageMath::fillInto(unsigned char ubFill, RasterImage *imgOut)
{
	ImageRect *outRect = imgOut->getValidRect();

	int nOutXo = outRect->getLeft(),
		nOutXn = outRect->getRight(),
		nOutYo = outRect->getTop(),
		nOutYn = outRect->getBottom();
		
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in maxInto(In,Out)");
                
	// for float output
	 if ( imgOut->hasFloatRaster() )
	 {
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		
		for (int nYi = nOutYo; nYi <= nOutYn; nYi++)
		{
			for (int nXi = nOutXo; nXi <= nOutXn; nXi++)
			{
				fOut[nYi][nXi] = (float)(ubFill);
			}
		}	
	 }
	 // for unsigned char output
	 else
	 {
		unsigned char** ubOut 	= ((RasterImage_gray*)imgOut)->getGray2D();
		
		for (int nYi = nOutYo; nYi <= nOutYn; nYi++)
		{
			for (int nXi = nOutXo; nXi <= nOutXn; nXi++)
			{
				ubOut[nYi][nXi] = ubFill;
			}
		}	
	 }

} // end ImageMath::fillInto(unsigned char ubFill, RasterImage *imgOut)

//-------------------------------

void ImageMath::fillInto(float fFill, RasterImage *imgOut)
{
	ImageRect *outRect = imgOut->getValidRect();

	int nOutXo = outRect->getLeft(),
		nOutXn = outRect->getRight(),
		nOutYo = outRect->getTop(),
		nOutYn = outRect->getBottom();
		
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in maxInto(In,Out)");
                
	// for float output
	 if ( imgOut->hasFloatRaster() )
	 {
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		
		for (int nYi = nOutYo; nYi <= nOutYn; nYi++)
		{
			for (int nXi = nOutXo; nXi <= nOutXn; nXi++)
			{
				fOut[nYi][nXi] = fFill;
			}
		}	
	 }
	 // for unsigned char output
	 else
	 {
	 	unsigned char ubFill = kWLC_Min((unsigned char)255,kWLC_Max((unsigned char)0,(unsigned char)fFill));
	 	
		unsigned char** ubOut 	= ((RasterImage_gray*)imgOut)->getGray2D();
		
		for (int nYi = nOutYo; nYi <= nOutYn; nYi++)
		{
			for (int nXi = nOutXo; nXi <= nOutXn; nXi++)
			{
				ubOut[nYi][nXi] = ubFill;
			}
		}	
	 }
	 	
} // end ImageMath::fillInto(float fFill, RasterImage *imgOut)

//-------------------------------

void ImageMath::fillUniformRandom(float fMinNoise, float fMaxNoise, RasterImage *imgOut)
{
	ImageRect *outRect = imgOut->getValidRect();

	int nOutXo = outRect->getLeft(),
		nOutXn = outRect->getRight(),
		nOutYo = outRect->getTop(),
		nOutYn = outRect->getBottom();
		
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in fillUniformRandom(fMin,fMax.Out)");
                
       
	// for float output
	 if ( imgOut->hasFloatRaster() )
	 {
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		
		for (int nYi = nOutYo; nYi <= nOutYn; nYi++)
		{
			for (int nXi = nOutXo; nXi <= nOutXn; nXi++)
			{
				fOut[nYi][nXi] = (fMaxNoise-fMinNoise)*rand()/RAND_MAX + fMinNoise;
			}
		}	
	 }
	 // for unsigned char output
	 else
	 {
	 	unsigned char ubFill;
	 	
		unsigned char** ubOut 	= ((RasterImage_gray*)imgOut)->getGray2D();
		
		for (int nYi = nOutYo; nYi <= nOutYn; nYi++)
		{
			for (int nXi = nOutXo; nXi <= nOutXn; nXi++)
			{
				float fNoise = (fMaxNoise-fMinNoise)*rand()/RAND_MAX + fMinNoise;
				if (fNoise < 0.0f)   ubFill=0;
				if (fNoise > 255.0f) ubFill=255;
				
				ubOut[nYi][nXi] = ubFill;
			}
		}	
	 }
	 	
} // end ImageMath::fillUniformRandom(float fMinNoise, float fMaxNoise, RasterImage *imgOut)

//-------------------------------

void ImageMath::fillUniformRandom(unsigned char ubMinNoise, unsigned char ubMaxNoise, RasterImage *imgOut)
{
	ImageRect *outRect = imgOut->getValidRect();

	int nOutXo = outRect->getLeft(),
		nOutXn = outRect->getRight(),
		nOutYo = outRect->getTop(),
		nOutYn = outRect->getBottom();
		
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in fillUniformRandom(fMin,fMax.Out)");
                
       
	// for float output
	 if ( imgOut->hasFloatRaster() )
	 {
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		
		for (int nYi = nOutYo; nYi <= nOutYn; nYi++)
		{
			for (int nXi = nOutXo; nXi <= nOutXn; nXi++)
			{
				fOut[nYi][nXi] = ((float)ubMaxNoise-(float)ubMinNoise)*rand()/RAND_MAX + (float)ubMinNoise;
			}
		}	
	 }
	 // for unsigned char output
	 else
	 {
	 	unsigned char ubFill;
	 	
		unsigned char** ubOut 	= ((RasterImage_gray*)imgOut)->getGray2D();
		
		for (int nYi = nOutYo; nYi <= nOutYn; nYi++)
		{
			for (int nXi = nOutXo; nXi <= nOutXn; nXi++)
			{
				float fNoise = ((float)ubMaxNoise-(float)ubMinNoise)*rand()/RAND_MAX + (float)ubMinNoise;
				if (fNoise < 0.0f)   ubFill=0;
				else if (fNoise > 255.0f) ubFill=255;
				else ubFill = (unsigned char)fNoise;
				
				ubOut[nYi][nXi] = ubFill;
			}
		}	
	 }
	 	
} // end ImageMath::fillUniformRandom(unsigned char fMinNoise, unsigned char fMaxNoise, RasterImage *imgOut)


//-------------------------------

void ImageMath::fillCheckerBoard(unsigned char ubBrightFill, unsigned char ubDarkFill,
								 unsigned char ubNumSquares, RasterImage *imgOut)
{
	ImageRect *outRect = imgOut->getValidRect();

	int nOutXo  = outRect->getLeft(),
		nOutXn  = outRect->getRight(),
		nOutYo  = outRect->getTop(),
		nOutYn  = outRect->getBottom(),
		nWidth  = outRect->getWidth(),
		nHeight = outRect->getHeight();
		
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in maxInto(In,Out)");
                
	// for float output
	 if ( imgOut->hasFloatRaster() )
	 {
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		
		for (int nYi = nOutYo; nYi <= nOutYn; nYi++)
		{
			int nCB_Row = nYi*ubNumSquares/nHeight;
			for (int nXi = nOutXo; nXi <= nOutXn; nXi++)
			{
				int nCB_Col = nXi*ubNumSquares/nWidth;
				// both even (or zero) DARK
				if (kWLC_IsEven(nCB_Row) && kWLC_IsEven(nCB_Col))
				{
					fOut[nYi][nXi] = (float)(ubDarkFill);
				}
				
				// even row odd col BRIGHT
				if (kWLC_IsEven(nCB_Row) && kWLC_IsOdd(nCB_Col))
				{
					fOut[nYi][nXi] = (float)(ubBrightFill);
				}
				
				// odd row even col BRIGHT
				if (kWLC_IsOdd(nCB_Row) && kWLC_IsEven(nCB_Col))
				{
					fOut[nYi][nXi] = (float)(ubBrightFill);
				}
				
				// odd row odd col DARK
				if (kWLC_IsOdd(nCB_Row) && kWLC_IsOdd(nCB_Col))
				{
					fOut[nYi][nXi] = (float)(ubDarkFill);
				}
			}
		}	
	 }
	 // for unsigned char output
	 else
	 {
		unsigned char** ubOut 	= ((RasterImage_gray*)imgOut)->getGray2D();
		
		for (int nYi = nOutYo; nYi <= nOutYn; nYi++)
		{
			int nCB_Row = nYi*ubNumSquares/nHeight;
			for (int nXi = nOutXo; nXi <= nOutXn; nXi++)
			{
				int nCB_Col = nXi*ubNumSquares/nWidth;
				// both even (or zero) DARK
				if (kWLC_IsEven(nCB_Row) && kWLC_IsEven(nCB_Col))
				{
					ubOut[nYi][nXi] = ubDarkFill;
				}
				
				// even row odd col BRIGHT
				if (kWLC_IsEven(nCB_Row) && kWLC_IsOdd(nCB_Col))
				{
					ubOut[nYi][nXi] = ubBrightFill;
				}
				
				// odd row even col BRIGHT
				if (kWLC_IsOdd(nCB_Row) && kWLC_IsEven(nCB_Col))
				{
					ubOut[nYi][nXi] = ubBrightFill;
				}
				
				// odd row odd col DARK
				if (kWLC_IsOdd(nCB_Row) && kWLC_IsOdd(nCB_Col))
				{
					ubOut[nYi][nXi] = ubDarkFill;
				}
			}
		}	
	 }

} // end ImageMath::fillCheckerBoard(ub bright, ub dark, num squares, image)

//-------------------------------

void ImageMath::fillCheckerBoard(float fBrightFill, float fDarkFill,
								 unsigned char ubNumSquares, RasterImage *imgOut)
{
	ImageRect *outRect = imgOut->getValidRect();

	int nOutXo  = outRect->getLeft(),
		nOutXn  = outRect->getRight(),
		nOutYo  = outRect->getTop(),
		nOutYn  = outRect->getBottom(),
		nWidth  = outRect->getWidth(),
		nHeight = outRect->getHeight();
		
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in maxInto(In,Out)");
                
	// for float output
	 if ( imgOut->hasFloatRaster() )
	 {
		float** fOut 	= ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		
		for (int nYi = nOutYo; nYi <= nOutYn; nYi++)
		{
			int nCB_Row = nYi*ubNumSquares/nHeight;
			for (int nXi = nOutXo; nXi <= nOutXn; nXi++)
			{
				int nCB_Col = nXi*ubNumSquares/nWidth;
				// both even (or zero) DARK
				if (kWLC_IsEven(nCB_Row) && kWLC_IsEven(nCB_Col))
				{
					fOut[nYi][nXi] = fDarkFill;
				}
				
				// even row odd col BRIGHT
				if (kWLC_IsEven(nCB_Row) && kWLC_IsOdd(nCB_Col))
				{
					fOut[nYi][nXi] = fBrightFill;
				}
				
				// odd row even col BRIGHT
				if (kWLC_IsOdd(nCB_Row) && kWLC_IsEven(nCB_Col))
				{
					fOut[nYi][nXi] = fDarkFill;
				}
				
				// odd row odd col DARK
				if (kWLC_IsOdd(nCB_Row) && kWLC_IsOdd(nCB_Col))
				{
					fOut[nYi][nXi] = fBrightFill;
				}
			}
		}	
	 }
	 // for unsigned char output
	 else
	 {
		unsigned char** ubOut 	= ((RasterImage_gray*)imgOut)->getGray2D();
		unsigned char ubDarkFill, ubBrightFill;
		
		// bound float fills to unsigned char limits
		if (fBrightFill < 0.0f)   ubBrightFill=0;
		if (fBrightFill > 255.0f) ubBrightFill=255;
		if (fDarkFill < 0.0f)     ubDarkFill=0;
		if (fDarkFill > 255.0f)   ubDarkFill=255;
		
		for (int nYi = nOutYo; nYi <= nOutYn; nYi++)
		{
			int nCB_Row = nYi*ubNumSquares/nHeight;
			for (int nXi = nOutXo; nXi <= nOutXn; nXi++)
			{
				int nCB_Col = nXi*ubNumSquares/nWidth;
				// both even (or zero) DARK
				if (kWLC_IsEven(nCB_Row) && kWLC_IsEven(nCB_Col))
				{
					ubOut[nYi][nXi] = ubDarkFill;
				}
				
				// even row odd col BRIGHT
				if (kWLC_IsEven(nCB_Row) && kWLC_IsOdd(nCB_Col))
				{
					ubOut[nYi][nXi] = ubBrightFill;
				}
				
				// odd row even col BRIGHT
				if (kWLC_IsOdd(nCB_Row) && kWLC_IsEven(nCB_Col))
				{
					ubOut[nYi][nXi] = ubBrightFill;
				}
				
				// odd row odd col DARK
				if (kWLC_IsOdd(nCB_Row) && kWLC_IsOdd(nCB_Col))
				{
					ubOut[nYi][nXi] = ubDarkFill;
				}
			}
		}	
	 }

} // end ImageMath::fillCheckerBoard(float bright, float dark, num squares, image)


//////////////////////////////////////////////////////////////////////////////////
//
// CIRCLE functions


void ImageMath::addCircleInto(float fXc, float fYc, float fRadius, float fFill, float fAngle, RasterImage *imgInOut)
{
	ImageRect *inOutRect = imgInOut->getValidRect();

	int nXo = inOutRect->getLeft(),
		nXn = inOutRect->getRight(),
		nYo = inOutRect->getTop(),
		nYn = inOutRect->getBottom();
		
	float 	fTheta = 0.0f, 
			fDeltaTheta = atan(1/fRadius);

	unsigned int unX, unY, 
				 unLastX = fXc - fRadius, 
				 unLastY = fYc;
		
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgInOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in fillCircleInto()");
                
	// for float output
	 if ( imgInOut->hasFloatRaster() )
	 {
		float** fOut = ((RasterImage_gray_F*)imgInOut)->getGrayF2D();
		
		while (fTheta<2*PI)
		{
			unX = fXc + fRadius*cos(fTheta);
			unY = fYc + fRadius*sin(fTheta)*cos(2.0f*PI*fAngle/360.0f);
			if ( (unX!=unLastX) || (unY!=unLastY) ) // it's new so add it in
			{
				unLastX = unX;
				unLastY = unY;
				if ( (nXo<=unX) && (unX<=nXn) && (nYo<=unY) && (unY<=nYn) )
				{
			   		fOut[unY][unX] += fFill;
				}
			}
			fTheta += fDeltaTheta;				
		}
	 }
	 // for unsigned char output
	 else
	 {
		unsigned char** ubInOut = ((RasterImage_gray*)imgInOut)->getGray2D();
	 	unsigned char ubFill = kWLC_Min((unsigned char)255,kWLC_Max((unsigned char)0,(unsigned char)fFill));
		
		while (fTheta<2*PI)
		{
			unX = fXc + fRadius*cos(fTheta);
			unY = fYc + fRadius*sin(fTheta)*cos(2.0f*PI*fAngle/360.0f);
			if ( (unX!=unLastX) || (unY!=unLastY) ) // it's new so add it in
			{
				unLastX = unX;
				unLastY = unY;
				if ( (nXo<=unX) && (unX<=nXn) && (nYo<=unY) && (unY<=nYn) )
				{
			   		ubInOut[unY][unX] = kWLC_Min(((unsigned char)255),((int)ubInOut[unY][unX])+((int)ubFill));
				}
			}
			fTheta += fDeltaTheta;				
		}
	 }

} // end ImageMath::addCircleInto()

//-------------------------------

void ImageMath::orCircleInto(float fXc, float fYc, float fRadius, float fFill, float fAngle, RasterImage *imgInOut)
{
	ImageRect *inOutRect = imgInOut->getValidRect();

	int nXo = inOutRect->getLeft(),
		nXn = inOutRect->getRight(),
		nYo = inOutRect->getTop(),
		nYn = inOutRect->getBottom();
		
	float 	fTheta = 0.0f, 
			fDeltaTheta = atan(1/fRadius);

	unsigned int unX, unY;
			
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgInOut->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in fillCircleInto()");
                
	// for float output
	 if ( imgInOut->hasFloatRaster() )
	 {
		float** fOut = ((RasterImage_gray_F*)imgInOut)->getGrayF2D();
		
		while (fTheta<2*PI)
		{
			unX = fXc + fRadius*cos(fTheta);
			unY = fYc + fRadius*sin(fTheta)*cos(2.0f*PI*fAngle/360.0f);
			if ( (nXo<=unX) && (unX<=nXn) && (nYo<=unY) && (unY<=nYn) )
			{
			   fOut[unY][unX] = fFill;
			}
			fTheta += fDeltaTheta;				
		}
	 }
	 // for unsigned char output
	 else
	 {
		unsigned char** ubInOut = ((RasterImage_gray*)imgInOut)->getGray2D();
	 	unsigned char ubFill = kWLC_Min((unsigned char)255,kWLC_Max((unsigned char)0,(unsigned char)fFill));
		
		while (fTheta<2*PI)
		{
			unX = fXc + fRadius*cos(fTheta);
			unY = fYc + fRadius*sin(fTheta)*cos(2.0f*PI*fAngle/360.0f);
			if ( (nXo<=unX) && (unX<=nXn) && (nYo<=unY) && (unY<=nYn) )
			{
			   ubInOut[unY][unX] = ubFill;
			}
			fTheta += fDeltaTheta;				
		}
	 }

} // end ImageMath::orCircleInto()



//////////////////////////////////////////////////////////////////////////////////
//
// STATS functions

void ImageMath::getStats(RasterImage *imgIn, ImageStats_Type *pStats)
{
	ImageRect *inRect = imgIn->getValidRect();

	int nOutXo = inRect->getLeft(),
		nOutXn = inRect->getRight(),
		nOutYo = inRect->getTop(),
		nOutYn = inRect->getBottom();
		
	int nYi, nXi;
		
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgIn->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in maxInto(In,Out)");
      
    // user inputs
    	// leave alone
    	  
    // init base stats  
    pStats->fMin =  LLONG_MAX;
    pStats->fMax = -LLONG_MAX;
    pStats->nMinX = -1;
    pStats->nMinY = -1;
    pStats->nMaxX = -1;
    pStats->nMaxY = -1;
    pStats->fSum = 0.0f;
    pStats->nCount = 0;          
    
    // set full stats to show they were not computed
    pStats->fSquaredSum = -1.0f;
    pStats->fSigma = -1.0f;
	pStats->nHistCount = -1;
          
	// for float output
	 if ( imgIn->hasFloatRaster() )
	 {
		float** fIn	= ((RasterImage_gray_F*)imgIn)->getGrayF2D();
		
		for (nYi = nOutYo; nYi <= nOutYn; nYi++)
		{
			for (nXi = nOutXo; nXi <= nOutXn; nXi++)
			{
				if (fIn[nYi][nXi]>pStats->fMax)
				{
					pStats->fMax = fIn[nYi][nXi];
					pStats->nMaxX = nXi;
					pStats->nMaxY = nYi;
				}
				if (fIn[nYi][nXi]<pStats->fMin)
				{
					pStats->fMin = fIn[nYi][nXi];
					pStats->nMinX = nXi;
					pStats->nMinY = nYi;
				}
				pStats->fSum += fIn[nYi][nXi];
				pStats->nCount++;
			}
		}
		pStats->fMean = pStats->fSum/pStats->nCount;
							
	 }
	 // for unsigned char output...
	 else
	 {
		unsigned char** ubIn = ((RasterImage_gray*)imgIn)->getGray2D();
		
		for (int nYi = nOutYo; nYi <= nOutYn; nYi++)
		{
			for (int nXi = nOutXo; nXi <= nOutXn; nXi++)
			{
				if (ubIn[nYi][nXi]>pStats->fMax)
				{
					pStats->fMax = ubIn[nYi][nXi];
					pStats->nMaxX = nXi;
					pStats->nMaxY = nYi;
				}
				if (ubIn[nYi][nXi]<pStats->fMin)
				{
					pStats->fMin = ubIn[nYi][nXi];
					pStats->nMinX = nXi;
					pStats->nMinY = nYi;
				}
				pStats->fSum += ubIn[nYi][nXi];
				pStats->nCount++;
			}
		}
		pStats->fMean = pStats->fSum/pStats->nCount;	
	 }
} // end ImageMath::getStats(RasterImage *imgIn, ImageStats_Type *pStats)

//-------------------------------

void ImageMath::getFullStats(RasterImage *imgIn, ImageStats_Type *pStats)
{
	ImageRect *inRect = imgIn->getValidRect();

	int nOutXo = inRect->getLeft(),
		nOutXn = inRect->getRight(),
		nOutYo = inRect->getTop(),
		nOutYn = inRect->getBottom();
		
	float fHistBin;
	
	int nYi, nXi, nBi;
		
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgIn->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in maxInto(In,Out)");
      
    // user inputs
    	// leave alone
    	  
    // init base stats  
    pStats->fMin =  LLONG_MAX;
    pStats->fMax = -LLONG_MAX;
    pStats->nMinX = -1;
    pStats->nMinY = -1;
    pStats->nMaxX = -1;
    pStats->nMaxY = -1;
    pStats->fSum = 0.0f;
    pStats->fSquaredSum = 0.0f;
    pStats->nCount = 0;          
    
    // set full stats to show they were not computed
    pStats->fSigma = 0.0f;
    pStats->nCount = 0;          
	pStats->nHistCount = 0;
          
	// for float output
	 if ( imgIn->hasFloatRaster() )
	 {
		float** fIn	= ((RasterImage_gray_F*)imgIn)->getGrayF2D();
		
		// first pass through the data
		
		for (nYi = nOutYo; nYi <= nOutYn; nYi++)
		{
			for (nXi = nOutXo; nXi <= nOutXn; nXi++)
			{
				if (fIn[nYi][nXi]>pStats->fMax)
				{
					pStats->fMax = fIn[nYi][nXi];
					pStats->nMaxX = nXi;
					pStats->nMaxY = nYi;
				}
				if (fIn[nYi][nXi]<pStats->fMin)
				{
					pStats->fMin = fIn[nYi][nXi];
					pStats->nMinX = nXi;
					pStats->nMinY = nYi;
				}
				pStats->fSum += fIn[nYi][nXi];
				pStats->fSquaredSum += (fIn[nYi][nXi])*(fIn[nYi][nXi]);
				pStats->nCount++;
			}
		}
		pStats->fMean = pStats->fSum/pStats->nCount;
		
		// second pass through the data

		pStats->nHistCount=0;
		if (pStats->nHistBins<1) pStats->nHistBins=0;
		if (pStats->nHistBins>MAX_HIST_BINS) pStats->nHistBins=MAX_HIST_BINS;
		for (nBi=0; nBi<MAX_HIST_BINS; nBi++) pStats->nBinCounts[nBi]=0;
		
		fHistBin = (pStats->fHistMax - pStats->fHistMin)/pStats->nHistBins;
		for (nYi = nOutYo; nYi <= nOutYn; nYi++)
		{
			for (nXi = nOutXo; nXi <= nOutXn; nXi++)
			{
				pStats->fSigma += pow(fIn[nYi][nXi]-pStats->fMean,2);
				for (nBi=1; nBi<=pStats->nHistBins; nBi++)
				{
					if (fIn[nYi][nXi]<=(pStats->fHistMin+nBi*fHistBin)) 
					{
						pStats->nBinCounts[nBi-1]++;
						pStats->nHistCount++;
						break;
					}
				}
			}
		}
		pStats->fSigma = sqrt(pStats->fSigma)/(pStats->nCount-1);
					
	 }
	 // for unsigned char output, ignore lots of the stats for now...
	 else
	 {
		unsigned char** ubIn = ((RasterImage_gray*)imgIn)->getGray2D();
		
		for (int nYi = nOutYo; nYi <= nOutYn; nYi++)
		{
			for (int nXi = nOutXo; nXi <= nOutXn; nXi++)
			{
				if (ubIn[nYi][nXi]<pStats->fMin)
				{
					pStats->fMin = ubIn[nYi][nXi];
					pStats->nMinX = nXi;
					pStats->nMinY = nYi;
				}
				pStats->fSum += ubIn[nYi][nXi];
				pStats->nCount++;
			}
		}
		pStats->fMean = pStats->fSum/pStats->nCount;	
	 }
} // end ImageMath::getFullStats(RasterImage *imgIn, ImageStats_Type *pStats)


//////////////////////////////////////////////////////////////////////////////////
//
// THRESHOLD functions

void ImageMath::thresholdInto(RasterImage *imgIn, float fPercent)
{
	ImageMath			  ImageFilter;
	ImageStats_Type		  ImageStats;

	ImageRect *inRect = imgIn->getValidRect();

	int nOutXo = inRect->getLeft(),
		nOutXn = inRect->getRight(),
		nOutYo = inRect->getTop(),
		nOutYn = inRect->getBottom();
		
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgIn->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in thresholdInto(In,fPercent)");

	// get stats for max value
	ImageFilter.getStats(imgIn, &ImageStats);

                
	// for float output
	 if ( imgIn->hasFloatRaster() )
	 {
		float** fIn = ((RasterImage_gray_F*)imgIn)->getGrayF2D();
		
		for (int nYi = nOutYo; nYi <= nOutYn; nYi++)
		{
			for (int nXi = nOutXo; nXi <= nOutXn; nXi++)
			{
				if (fIn[nYi][nXi] < fPercent*ImageStats.fMax)
				{
					fIn[nYi][nXi] = 0.0f;
				}
			}
		}	
	 }
	 // for unsigned char output
	 else
	 {
		unsigned char** ubIn = ((RasterImage_gray*)imgIn)->getGray2D();
		
		for (int nYi = nOutYo; nYi <= nOutYn; nYi++)
		{
			for (int nXi = nOutXo; nXi <= nOutXn; nXi++)
			{
				if (ubIn[nYi][nXi] < fPercent*ImageStats.fMax)
				{
					ubIn[nYi][nXi] = 0.0f;
				}
			}
		}	
	 }
	 	
} // end ImageMath::thresholdInto(RasterImage *imgIn, float fPercent)

//-------------------------------

void ImageMath::threshold(RasterImage *imgIn, float fPercent, RasterImage *imgOut)
{
	ImageMath			  ImageFilter;
	ImageStats_Type		  ImageStats;

	ImageRect *inRect = imgIn->getValidRect();

	int nOutXo = inRect->getLeft(),
		nOutXn = inRect->getRight(),
		nOutYo = inRect->getTop(),
		nOutYn = inRect->getBottom();
		
	// make sure we have a gray level image and not color
	FAIL_CONDITION(imgIn->getBaseRasterType() != kGrayRaster,
                kRasterImageError,
                "Only gray-level output rasters accepted in thresholdInto(In,fPercent)");

	// get stats for max value
	ImageFilter.getStats(imgIn, &ImageStats);

                
	// for float output
	 if ( imgIn->hasFloatRaster() && imgOut->hasFloatRaster() )
	 {
		float** fIn = ((RasterImage_gray_F*)imgIn)->getGrayF2D();
		float** fOut = ((RasterImage_gray_F*)imgOut)->getGrayF2D();
		
		for (int nYi = nOutYo; nYi <= nOutYn; nYi++)
		{
			for (int nXi = nOutXo; nXi <= nOutXn; nXi++)
			{
				if (fIn[nYi][nXi] < fPercent*ImageStats.fMax)
				{
					fOut[nYi][nXi] = 0.0f;
				}
				else
				{
					fOut[nYi][nXi] = fIn[nYi][nXi];
				}
			}
		}	
	 }
	 // for unsigned char output
	 else
	 {
		unsigned char** ubIn 	= ((RasterImage_gray*)imgIn)->getGray2D();
		unsigned char** ubOut 	= ((RasterImage_gray*)imgOut)->getGray2D();
		
		for (int nYi = nOutYo; nYi <= nOutYn; nYi++)
		{
			for (int nXi = nOutXo; nXi <= nOutXn; nXi++)
			{
				if (ubIn[nYi][nXi] < fPercent*ImageStats.fMax)
				{
					ubOut[nYi][nXi] = 0.0f;
				}
				else
				{
					ubOut[nYi][nXi] = ubIn[nYi][nXi];
				}
			}
		}	
	 }
	 
	 imgOut->setValidRect(imgIn->getValidRect());
	 	
} // end ImageMath::threshold(RasterImage *imgIn, float fPercent, RasterImage *imgOut)

