#ifndef IMAGE_MATH_H
#define IMAGE_MATH_H

#include "collier_CSC583LIB.h"

using namespace csc583Base;

#define MAX_HIST_BINS 20

typedef struct
{
	// user inputs
	int   nHistBins;
	float fHistMin;
	float fHistMax;
	// quick stats
	float fMin;
	int   nMinX, nMinY;
	float fMean;
	float fMax;
	int   nMaxX, nMaxY;
	float fSum;
	int   nCount;
	//full stats
	float fSquaredSum;
	float fSigma;
	int	  nHistCount;
	int   nBinCounts[MAX_HIST_BINS];
} ImageStats_Type;

//
// function declarations
//

namespace csc583ImageProcessing
{

    class CSC583LIB_EXPORT ImageMath
    {
        public:

            /** Constructs the basic image math object
             */
            ImageMath(void);

          
            /** Destructor
             */
            ~ImageMath(void);
            
             /** Copies the first RasterImage to the second. No conversions are done.
             *   The copy is made to the same X,Y space in imgOut.
             *
             *  @param  imgIn   the RasterImage to be copied
             *  @param  imgOut  the RasterImage into which the output is written
             */ 

          void copy(RasterImage *imgIn, RasterImage *imgOut);
            
             /** Copies the first RasterImage to the second. No conversions are done.
             *   The copy is made to the X,Y space in imgOut at the origin nXo, nYo.
             *
             *  @param  imgIn   the RasterImage to be copied
             *  @param  nXo     destination X origin
             *  @param  nYo     destination Y origin
             *  @param  imgOut  the RasterImage into which the output is written
             */ 

          void copyTo(RasterImage *imgIn, int nXo, int nYo, RasterImage *imgOut);
          
             /** Copies the first RasterImage to the second but the pixel values are
             *   flipped within s.t. min--> max and max-->min.
             *   The copy is made to the same X,Y space in imgOut.
             *
             *  @param  imgIn   the RasterImage to be copied
             *  @param  imgOut  the RasterImage into which the output is written
             */ 

          void copyFlip(RasterImage *imgIn, RasterImage *imgOut);

             /** Transposes the raster in place.
             *
             *  @param  imgIn   the RasterImage to be transposed
             */ 

          void transpose(RasterImage *imgIn);

            /** Correlates the first RasterImage to the second and writes the result
             *	into the third RasterImage.
             *
             *  @param  imgA       the RasterImage to which the operator is applied
             *  @param  imgB       the RasterImage to which the operator is applied
             *  @param  imgOut     the RasterImage into which the output is written
             */ 

          void correlate(RasterImage *imgA, RasterImage *imgB, RasterImage *imgOut);

            /** Convolves the first RasterImage to the second and writes the result
             *	into the third RasterImage.
             *
             *  @param  imgA       the RasterImage to which the operator is applied
             *  @param  imgB       the RasterImage to which the operator is applied
             *  @param  imgOut     the RasterImage into which the output is written
             */ 

          void convolve(RasterImage *imgA, RasterImage *imgB, RasterImage *imgOut);
          
             /** multiplies the first RasterImage to the second and writes the result
             *	into the third RasterImage.
             *
             *  @param  imgA       the RasterImage to which the operator is applied
             *  @param  imgB       the RasterImage to which the operator is applied
             *  @param  imgOut     the RasterImage into which the output is written
             */ 

          void multiply(RasterImage *imgA, RasterImage *imgB, RasterImage *imgOut);

             /** multiplies the first RasterImage to the second and writes the result
             *	into the third RasterImage. Treats the inputs as complex so the
             *  output is complex as well.
             *
             *  @param  imgA       the RasterImage to which the operator is applied
             *  @param  imgB       the RasterImage to which the operator is applied
             *  @param  imgOut     the RasterImage into which the output is written
             */ 

          void multiplyComplex(RasterImage *imgA, RasterImage *imgB, RasterImage *imgOut);

             /** scales the first RasterImage 
             *
             *  @param  imgA       the RasterImage to which the operator is applied
             *  @param  fScale     the scale factor
             */ 

          void scale(RasterImage *imgA, float fScale);

             /** scales the first RasterImage and writes the result
             *	into the second RasterImage.
             *
             *  @param  imgA       the RasterImage to which the operator is applied
             *  @param  fScale     the scale factor
             *  @param  imgOut     the RasterImage into which the output is written
             */ 

          void scale(RasterImage *imgA, float fScale, RasterImage *imgOut);
          
            /** Adds the first RasterImage into the second and writes the result
             *	into the second RasterImage.
             *
             *  @param  imgIn       the RasterImage to which the operator is applied
             *  @param  imgOut      the RasterImage into which the output is written
             */ 

          void addInto(RasterImage *imgIn, RasterImage *imgOut);
            
            /** Adds the first RasterImage to the second and writes the result
             *	into the third RasterImage.
             *
             *  @param  imgA       the RasterImage to which the operator is applied
             *  @param  imgB       the RasterImage to which the operator is applied
             *  @param  imgOut     the RasterImage into which the output is written
             */ 

          void add(RasterImage *imgA, RasterImage *imgB, RasterImage *imgOut);

            /** Subtracts the first RasterImage from the second and writes the result
             *	back into the second RasterImage.
             *
             *  @param  imgIn       the RasterImage to which the operator is applied
             *  @param  imgOut      the RasterImage into which the output is written
             */ 

          void subtractFrom(RasterImage *imgIn, RasterImage *imgOut);
            
            /** Subtracts the first RasterImage from the second and writes the result
             *	into the third RasterImage.
             *
             *  @param  imgA       the RasterImage to which the operator is applied
             *  @param  imgB       the RasterImage to which the operator is applied
             *  @param  imgOut     the RasterImage into which the output is written
             */ 

          void subtract(RasterImage *imgA, RasterImage *imgB, RasterImage *imgOut);

            /** Finds the max of the first RasterImage and the second and writes the result
             *	back into the second RasterImage.
             *
             *  @param  imgIn       the RasterImage to which the operator is applied
             *  @param  imgOut      the RasterImage into which the output is written
             */ 

          void maxInto(RasterImage *imgIn, RasterImage *imgOut);
            
            /** Finds the max of the first RasterImage and the second and writes the result
             *	into the third RasterImage.
             *
             *  @param  imgA       the RasterImage to which the operator is applied
             *  @param  imgB       the RasterImage to which the operator is applied
             *  @param  imgOut     the RasterImage into which the output is written
             */ 

          void maxOf(RasterImage *imgA, RasterImage *imgB, RasterImage *imgOut);

            /** Combines nScale pixels of the first RasterImage into the second 
             *
             *  @param  imgA       the RasterImage to which the operator is applied
             *  @param  nScale     number of pixels to combine in each direction
             *  @param  imgOut     the RasterImage into which the output is written
             */ 
		  void combine(RasterImage *imgA, int nScale, RasterImage *imgOut);

            /** Uncombines nScale pixels of the first RasterImage into the second 
             *
             *  @param  imgA       the RasterImage to which the operator is applied
             *  @param  nScale     number of pixels to combine in each direction
             *  @param  imgOut     the RasterImage into which the output is written
             */ 
		  void uncombine(RasterImage *imgA, int nScale, RasterImage *imgOut);

            /** Copies a region of the first RasterImage into the second 
             *
             *  @param  imgA       the RasterImage to which the operator is applied
             *  @param  nXc        desired X center
             *  @param  nYc        desired Y center
             *  @param  imgOut     the RasterImage into which the output is written
             */ 

		  void selectRegion(RasterImage *imgA, int nXc, int nYc, RasterImage *imgOut);

            /** Fills the imgOut raster with the specified data
             *
             *  @param  ubFill      the data to fill the image with
             *  @param  imgOut      the RasterImage into which the output is written
             */ 

          void fillInto(unsigned char ubFill, RasterImage *imgOut);
          

           /** Fills the imgOut raster with the specified data
             *
             *  @param  fFill      the data to fill the image with
             *  @param  imgOut      the RasterImage into which the output is written
             */ 

          void fillInto(float fFill, RasterImage *imgOut);

            /** Fills the imgOut raster with uniform random noise
             *	based on the min/max parameters.
             *
             *  @param  fMinNoise   min noise value
             *  @param  fMaxNoise   max noise value
             *  @param  imgOut      the RasterImage into which the output is written
             */ 

          void fillUniformRandom(float fMinNoise, float fMaxNoise, RasterImage *imgOut);

            /** Fills the imgOut raster with uniform random noise
             *	based on the min/max parameters.
             *
             *  @param  ubMinNoise   min noise value
             *  @param  ubMaxNoise   max noise value
             *  @param  imgOut      the RasterImage into which the output is written
             */ 

          void fillUniformRandom(unsigned char fMinNoise, unsigned char fMaxNoise, RasterImage *imgOut);

            /** Fills the imgOut raster with a checkerboard pattern
             *	as specified by the parameters.
             *
             *  @param  ubBrightFill  bright square gray level
             *  @param  ubDarkFill    dark square gray level
             *  @param  ubNumSquares  number of squares to generate along each axis 
             *  @param  imgOut        the RasterImage into which the output is written
             */ 

		  void fillCheckerBoard(unsigned char ubBrightFill, unsigned char ubDarkFill,
								 unsigned char ubNumSquares, RasterImage *imgOut);
								 

             /** Fills the imgOut raster with a checkerboard pattern
             *	as specified by the parameters.
             *
             *  @param  fBrightFill  bright square gray level
             *  @param  fDarkFill    dark square gray level
             *  @param  ubNumSquares  number of squares to generate along each axis 
             *  @param  imgOut        the RasterImage into which the output is written
             */ 

		  void fillCheckerBoard(float ubBrightFill, float ubDarkFill,
								 unsigned char ubNumSquares, RasterImage *imgOut);

            /** Fills the imgInOut raster with a circle based upon the
             *	specified Xo, Yo, and Radius. The specified points are set to the
             * specified value. All other points are left unchanged.
             *
             *  @param  fXc   		the circle's center X
             *  @param  fYc   		the circle's center Y
             *  @param  fRadius    	the circle's radius
             *  @param  fFill     	the value to set the circle points to
             *  @param  fAngle     	the angle to rotate the circle around H axis
             *  @param  imgInOut    the RasterImage into which the output is written
             */ 

          void addCircleInto(float fXc, float fYc, float fRadius, float fFill, float fAngle, RasterImage *imgInOut);

            /** Fills the imgInOut raster with a circle based upon the
             *	specified Xo, Yo, and Radius. The specified points are set to the
             * specified value. All other points are left unchanged.
             *
             *  @param  fXc   		the circle's center X
             *  @param  fYc   		the circle's center Y
             *  @param  fRadius    	the circle's radius
             *  @param  fFill     	the value to set the circle points to
             *  @param  fAngle     	the angle to rotate the circle around H axis
             *  @param  imgInOut    the RasterImage into which the output is written
             */ 

           void orCircleInto(float fXc, float fYc, float fRadius, float fFill, float fAngle, RasterImage *imgInOut);

           /** Fills the stats structure with the specified data
             *
             *  @param  imgIn       the RasterImage to get stats on
             *  @param  pStats      the statistics of the image
             */ 

          void getStats(RasterImage *imgIn, ImageStats_Type *pStats);

           /** Fills the stats structure with the specified data
             *
             *  @param  imgIn       the RasterImage to get stats on
             *  @param  pStats      the statistics of the image
             */ 

          void getFullStats(RasterImage *imgIn, ImageStats_Type *pStats);

             /** Thresholds imgIn into itself.
             *
             *  @param  imgIn       the RasterImage to which the operator is applied
             *  @param  fPercent    the threshold percent
             */ 

          void thresholdInto(RasterImage *imgIn, float fPercent);
            
            /** Thresholds imgIn into imgOut.
             *
             *  @param  imgIn       the RasterImage to which the operator is applied
             *  @param  fPercent    the threshold percent
             *  @param  imgOut      the RasterImage into which the output is written
             */ 

          void threshold(RasterImage *imgIn, float fPercent, RasterImage *imgOut);
          
       protected:


    };
}
#endif  //  IMAGE_MATH_H